from metabolights_utils.provider.ftp import (
    default_ftp_client,
    folder_metadata_collector,
    model,
)

__all__ = [
    "default_ftp_client",
    "folder_metadata_collector",
    "model",
]



from functools import lru_cache

from metabolights_utils.isatab.default.factory import (
    DefaultIsaTabReaderFactory,
    DefaultIsaTabWriterFactory,
)
from metabolights_utils.isatab.reader import (
    InvestigationFileReader,
    IsaTableFileReader,
    IsaTabReaderFactory,
)
from metabolights_utils.isatab.writer import (
    InvestigationFileWriter,
    IsaTableFileWriter,
    IsaTabWriterFactory,
)

from metabolights_utils.isatab import default, reader, writer


@lru_cache
def get_reader_factory():
    reader_factory: IsaTabReaderFactory = DefaultIsaTabReaderFactory()
    return reader_factory


@lru_cache
def get_writer_factory():
    writer_factory: IsaTabWriterFactory = DefaultIsaTabWriterFactory()
    return writer_factory


class Reader:
    reader_factory: IsaTabReaderFactory = get_reader_factory()

    @classmethod
    def get_investigation_file_reader(cls) -> InvestigationFileReader:
        return cls.reader_factory.get_investigation_file_reader()

    @classmethod
    def get_assay_file_reader(cls, results_per_page=100) -> IsaTableFileReader:
        return cls.reader_factory.get_assay_file_reader(
            results_per_page=results_per_page
        )

    @classmethod
    def get_sample_file_reader(cls, results_per_page=100) -> IsaTableFileReader:
        return cls.reader_factory.get_assay_file_reader(
            results_per_page=results_per_page
        )

    @classmethod
    def get_assignment_file_reader(cls, results_per_page=100) -> IsaTableFileReader:
        return cls.reader_factory.get_assay_file_reader(
            results_per_page=results_per_page
        )


class Writer:
    writer_factory: IsaTabWriterFactory = get_writer_factory()

    @classmethod
    def get_investigation_file_writer(cls) -> InvestigationFileWriter:
        return cls.writer_factory.get_investigation_file_writer()

    @classmethod
    def get_assay_file_writer(cls) -> IsaTableFileWriter:
        return cls.writer_factory.get_assay_file_writer()

    @classmethod
    def get_sample_file_writer(cls) -> IsaTableFileWriter:
        return cls.writer_factory.get_sample_file_writer()

    @classmethod
    def get_assignment_file_writer(cls) -> IsaTableFileWriter:
        return cls.writer_factory.get_sample_file_writer()


__all__ = [
    "reader",
    "default",
    "writer",
    "Reader",
    "Writer",
    "get_reader_factory",
    "get_writer_factory",
]



import os
import shutil
from typing import Union

import click

from metabolights_utils.provider import definitions


@click.command(name="remove", no_args_is_help=True)
@click.option(
    "--local_path",
    "-p",
    default=definitions.default_local_repority_root_path,
    help="Local storage root path. Folder will be created if it does not exist.",
)
@click.option(
    "--local_cache_path",
    "-c",
    default=definitions.default_local_repority_cache_path,
    help="Path to store cache files of FTP file indices, study models, etc.",
)
@click.argument("study_id")
def public_remove(
    study_id: Union[None, str] = None,
    local_path: Union[None, str] = None,
    local_cache_path: Union[None, str] = None,
):
    """
    Delete local study data and metadata files.

    study_id: MetaboLights study accession number (MTBLSxxxx).
    """

    study_id = study_id.upper()
    try:
        data_path = os.path.join(local_path, study_id)
        cache_path = os.path.join(local_cache_path, study_id)
        deleted = False
        for folder in (data_path, cache_path):
            if os.path.exists(folder):
                shutil.rmtree(folder)
                deleted = True
        if deleted:
            click.echo(f"Study {study_id} folders are deleted.")
        else:
            click.echo(f"There is no folder for study {study_id}.")
    except Exception as ex:
        click.echo(f"Remove study {study_id} error: {ex}.")
        exit(1)


if __name__ == "__main__":
    public_remove(["MTBLS1"])



import datetime
from enum import Enum
from typing import Any, Dict, Generic, List, TypeVar, Union
from typing_extensions import Annotated

from pydantic import BaseModel, ConfigDict, Field

from metabolights_utils.common import CamelCaseModel


class ValidationMessage(BaseModel):
    message: str = ""
    section: str = ""
    val_sequence: str = ""
    status: str = ""
    metadata_file: str = ""
    value: str = ""
    description: str = ""
    val_override: str = ""
    val_message: str = ""
    comment: str = ""


class Section(BaseModel):
    section: str = ""
    has_more_items: bool = False
    details: List[ValidationMessage] = []


class Validation(BaseModel):
    version: str = ""
    status: str = ""
    timig: Union[float, int] = 0.0
    last_update_time: str = ""
    last_update_timestamp: Union[float, int] = 0
    validations: List[Section] = []


class ValidationReport(BaseModel):
    validation: Validation = Validation()


class ValidationTaskStatus(BaseModel):
    task_id: str = ""
    last_update_time: float = 0
    last_update_timestamp: str = ""
    last_status: str = ""
    task_done_time: float = 0
    task_done_time_str: str = ""


class ValidationResponse(BaseModel):
    new_task: bool = False
    message: str = ""
    task: ValidationTaskStatus = ValidationTaskStatus()


class ValidationResponseV2(CamelCaseModel):
    status: bool = False
    status_message: str = ""
    task: ValidationTaskStatus = ValidationTaskStatus()


T = TypeVar("T", bound=CamelCaseModel)

LIST_ITEM = TypeVar("LIST_ITEM")


class Status(str, Enum):
    SUCCESS = "success"
    ERROR = "error"


class APIValidationError(CamelCaseModel):
    type: Annotated[
        str,
        Field(description="Validation error type"),
    ] = ""

    loc: Annotated[
        Union[str, List[Union[int, str]]],
        Field(description="Location of the error"),
    ] = ""
    msg: Annotated[
        str,
        Field(description="Error message"),
    ] = ""

    input: Annotated[
        Union[str, Dict[str, Any]],
        Field(description="Input data that caused the error"),
    ] = ""


class APIBaseResponse(CamelCaseModel):
    status: Annotated[
        Status,
        Field(description="Status of the response. It can be `success` or `error`"),
    ] = Status.SUCCESS

    success_message: Annotated[
        Union[None, str],
        Field(
            description="If status is `success`, response may contain success message."
            "It may contain a message for partial success even if response status is `error`."
        ),
    ] = None

    error_message: Annotated[
        Union[None, str],
        Field(description="If status is `error`, response may contain error message."),
    ] = None

    errors: Annotated[
        Union[List[str], List[APIValidationError]],
        Field(
            description="If status is `error`, response may contain a list of error details."
        ),
    ] = []


class APIResponse(APIBaseResponse, Generic[T]):
    """
    API response model for non-paginated results.
    """

    content: Annotated[
        Union[None, T],
        Field(description="If status is `success`, this stores response data."),
    ] = None


class WorkerTaskStatus(BaseModel):
    task_id: Annotated[
        str,
        Field(
            description="This field contains task id of the task.",
        ),
    ] = ""
    task_status: Annotated[
        str,
        Field(
            description="This field contains status of the task. Values: INITIATED, STARTED, SUCCESS, FAILURE, REVOKED, PENDING, etc.",
        ),
    ] = ""
    message: Annotated[
        str,
        Field(
            description="Message related to the task",
        ),
    ] = ""


class PolicyMessageType(str, Enum):
    ERROR = "ERROR"
    WARNING = "WARNING"
    INFO = "INFO"
    SUCCESS = "SUCCESS"


class PolicyMessage(CamelCaseModel):
    identifier: str = ""
    title: str = ""
    description: str = ""
    section: str = ""
    priority: str = ""
    type: PolicyMessageType = PolicyMessageType.INFO
    violation: str = ""
    source_file: str = ""
    source_column_header: str = ""
    source_column_index: Union[str, int] = ""
    has_more_miolations: bool = False
    total_violations: int = 0
    values: List[str] = []
    technique: str = ""


class OpaValidationResult(BaseModel):
    violations: Annotated[List[PolicyMessage], Field()] = []
    summary: Annotated[List[PolicyMessage], Field([])] = []


class PolicySummaryResult(BaseModel):
    study_id: str = ""
    start_time: Union[None, str, datetime.datetime] = None
    completion_time: Union[None, str, datetime.datetime] = None
    duration_in_seconds: float = 0
    messages: OpaValidationResult = OpaValidationResult()
    metadata_updates: List[str] = []
    metadata_modifier_enabled: bool = False
    assay_file_techniques: Dict[str, str] = {}
    maf_file_techniques: Dict[str, str] = {}


class PolicyResultResponse(WorkerTaskStatus):
    task_result: Annotated[
        PolicySummaryResult,
        Field(
            description="This field contains task id of the task.",
        ),
    ] = PolicySummaryResult()


class FtpUploadDetails(BaseModel):
    study_id: str
    ftp_folder: str
    ftp_host: str
    ftp_user: str
    ftp_password: str



import glob
import logging
import os
from typing import List

logger = logging.getLogger(__name__)

from metabolights_utils.utils.filename_utils import join_path


class MetabolightsSearchUtils(object):
    @staticmethod
    def get_isa_metadata_files(folder_path: str, recursive: bool = False) -> List[str]:
        metadata_files = []
        patterns = ["[asi]_*.txt", "m_*.tsv"]
        if not os.path.exists(folder_path) or not os.path.isdir(folder_path):
            logger.warning("Not valid directory %s", folder_path)
            return metadata_files

        for pattern in patterns:
            search_pattern = join_path(folder_path, pattern)
            metadata_files.extend(glob.glob(search_pattern, recursive=recursive))
        return metadata_files



import logging

from metabolights_utils.isatab.default.assay_file import DefaultAssayFileReader
from metabolights_utils.isatab.default.assignment_file import (
    DefaultAssignmentFileReader,
)
from metabolights_utils.isatab.default.investigation_file import (
    DefaultInvestigationFileReader,
    DefaultInvestigationFileWriter,
)
from metabolights_utils.isatab.default.sample_file import DefaultSampleFileReader
from metabolights_utils.isatab.default.writer import DefaultIsaTableFileWriter
from metabolights_utils.isatab.reader import (
    InvestigationFileReader,
    IsaTableFileReader,
    IsaTabReaderFactory,
)
from metabolights_utils.isatab.writer import (
    InvestigationFileWriter,
    IsaTableFileWriter,
    IsaTabWriterFactory,
)

logger = logging.getLogger(__name__)


class DefaultIsaTabReaderFactory(IsaTabReaderFactory):
    def get_investigation_file_reader(self) -> InvestigationFileReader:
        logger.debug("Return default investigation file reader.")
        return DefaultInvestigationFileReader()

    def get_assay_file_reader(self, results_per_page=100) -> IsaTableFileReader:
        logger.debug("Return default assay file reader.")
        return DefaultAssayFileReader(results_per_page=results_per_page)

    def get_sample_file_reader(self, results_per_page=100) -> IsaTableFileReader:
        logger.debug("Return default sample file reader.")
        return DefaultSampleFileReader(results_per_page=results_per_page)

    def get_assignment_file_reader(self, results_per_page=100) -> IsaTableFileReader:
        logger.debug("Return default maf file reader.")
        return DefaultAssignmentFileReader(results_per_page=results_per_page)


class DefaultIsaTabWriterFactory(IsaTabWriterFactory):
    def get_investigation_file_writer(self) -> InvestigationFileWriter:
        logger.debug("Return default investigation file writer.")
        return DefaultInvestigationFileWriter()

    def get_assay_file_writer(self) -> IsaTableFileWriter:
        logger.debug("Return default ISA table file writer for assay file.")
        return DefaultIsaTableFileWriter()

    def get_sample_file_writer(self) -> IsaTableFileWriter:
        logger.debug("Return default ISA table file writer for sample file.")
        return DefaultIsaTableFileWriter()

    def get_assignment_file_writer(self) -> IsaTableFileWriter:
        logger.debug("Return default ISA table file writer for maf file.")
        return DefaultIsaTableFileWriter()



from typing import List

from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.models.isa.common import AssayTechnique, IsaTableFile


class AssayFile(IsaTableFile):
    sample_names: Annotated[
        List[str],
        Field(description="Unique values in `Sample Name` column in the assay file."),
    ] = []

    assay_names: Annotated[
        List[str],
        Field(description="Unique values in `Assay Name` column in the assay file."),
    ] = []

    referenced_assignment_files: Annotated[
        List[str],
        Field(
            description="Relative paths of Metabolite Assignment File in the assay file."
        ),
    ] = []

    referenced_raw_files: Annotated[
        List[str],
        Field(description="Relative paths of referenced raw files in the assay file."),
    ] = []

    referenced_derived_files: Annotated[
        List[str],
        Field(
            description="Relative paths of referenced derived files in the assay file."
        ),
    ] = []

    referenced_raw_file_extensions: Annotated[
        List[str],
        Field(
            description="Unique file extensions of referenced raw files in the assay file."
        ),
    ] = []

    referenced_derived_file_extensions: Annotated[
        List[str],
        Field(
            description="Unique file extensions of referenced derived files in the assay file."
        ),
    ] = []

    assay_technique: Annotated[
        AssayTechnique, Field(description="Assay technique information.")
    ] = AssayTechnique()

    number_of_assay_rows: Annotated[
        int,
        Field(description="Number of assay rows."),
    ] = 0



from metabolights_utils.tsv.actions import (
    add_column,
    add_row,
    base,
    copy_column,
    copy_row,
    delete_column,
    delete_row,
    move_column,
    move_row,
    update_cell,
    update_column_header,
    update_column,
    update_row,
)

__all__ = [
    "add_column",
    "add_row",
    "base",
    "copy_column",
    "copy_row",
    "delete_column",
    "delete_row",
    "move_column",
    "move_row",
    "update_cell",
    "update_column_header",
    "update_column",
    "update_row",
]



import json
from typing import Any, Callable, Dict, Literal, Union

import click

from metabolights_utils.commands.utils import convert_html_to_plain_text, split_to_lines
from metabolights_utils.provider import definitions
from metabolights_utils.provider.utils import rest_api_post

JoinOperator = Literal["and", "or"]


@click.command(no_args_is_help=True, name="search")
@click.option(
    "--search_rest_api_url",
    "-u",
    default=definitions.default_study_search_rest_api_url,
    help="MetaboLights search API URL.",
)
@click.option(
    "--skip",
    "-s",
    default=0,
    type=int,
    help="Skip n items from the matched items.",
)
@click.option(
    "--limit",
    "-l",
    default=10,
    type=int,
    help="Maximum number items in response. Maximum return size is 100 items.",
)
@click.option(
    "--query_join_operator",
    "-j",
    default="and",
    type=str,
    help="If multiple keywords are defined and there is no join operator (+, |) in query, One of the 'and' (default) or 'or' operator will be used.",
)
@click.option(
    "--body",
    "-b",
    default=None,
    type=str,
    help="Advanced filter options in json format. Please read the API documentation.",
)
@click.option(
    "--study_ids",
    "--id",
    is_flag=True,
    default=False,
    help="Shows only MetaboLights accession numbers.",
)
@click.option(
    "--raw",
    is_flag=True,
    default=False,
    help="Shows raw result in json format.",
)
@click.argument("query", required=False)
def public_search(
    search_rest_api_url: str = "",
    skip: int = 0,
    limit: int = 10,
    query_join_operator: JoinOperator = "and",
    body: Union[None, str] = None,
    query: Union[None, str] = None,
    study_ids: bool = False,
    raw: bool = False,
):
    """
    Search public studies with query keywords. If there are multiple search keywords and no join operator (+, |) defined, results are merged with the selected query join operator (and, or)

    query: query terms that will be in search. e.g. cancer, (mus musculus)
    """

    sub_path = "/public/search/studies/_search"
    parameters = {
        "limit": limit,
        "skip": skip,
        "query_join_operator": query_join_operator,
    }
    if query:
        parameters["query"] = query.strip()
    else:
        query = ""
    try:
        body = json.loads(body) if body else {}
    except Exception as ex:
        click.echo(str(ex))
        exit(1)
    headers = {}
    url = f"{search_rest_api_url.rstrip('/')}/{sub_path.lstrip('/')}"
    response, error = rest_api_post(
        url=url,
        headers=headers,
        parameters=parameters,
        json_body=body,
    )
    if response and not error:
        if raw and "content" in response:
            str_response = json.dumps(response["content"], indent=4)
            click.echo(str_response)
        else:
            print_search_response(study_ids, query, response, click.echo)
    else:
        click.echo(error)
        exit(1)


def print_search_response(
    show_only_study_ids: bool, query: str, response: Dict[str, Any], log: Callable
):
    if "status" in response and response["status"].lower() == "success":
        log(f"MetaboLights public study search results for query: {query}")
    else:
        log(f"Failure of MetaboLights public study search for query: {query}")
        exit(1)

    if (
        "content" not in response
        or not response["content"]
        or "page" not in response["content"]
        or "total" not in response["content"]
        or "pageSize" not in response["content"]
    ):
        log("Invalid response.")
        exit(1)

    content = response["content"]
    page = content["page"]
    total = content["total"]
    page_size = content["pageSize"]

    if total > 0:
        if show_only_study_ids:
            id_list = [x["studyId"] for x in page if "studyId" in x]
            log(f"[{', '.join(id_list)}]")
            exit(0)
        else:
            log(f"{page_size} of total {total} results")
    else:
        if show_only_study_ids:
            log("[]")
        else:
            log("No results found.")
        exit(0)

    for item in page:
        if "studyId" in item and "title" in item:
            log(f"  - {item['studyId']}: {item['title']}")
        if "contacts" in item:
            names = [f'{x["firstName"]} {x["lastName"]}' for x in item["contacts"]]
            log(f"\t* Submitters: {', '.join(names)}")

        if "description" in item:
            log("\t* Description:")
            lines = split_to_lines(convert_html_to_plain_text(item["description"]))
            for line in lines:
                log(f"\t  {line}")
        if "assayTechniques" in item:
            names = [
                x["name"] for x in item["assayTechniques"] if "name" in x and x["name"]
            ]
            log(f"\t* Techniques: {', '.join(names).strip(',').strip().strip(',')}")
        for terms in [
            ("designDescriptors", "Design descriptors"),
            ("factors", "Factors"),
            ("organisms", "Organisms"),
            ("organismParts", "Organism parts"),
        ]:
            if terms[0] in item:
                log(f"\t* {terms[1]}:")
                for descriptor in item[terms[0]]:
                    parts = [
                        descriptor["termSourceRef"],
                        descriptor["term"],
                        descriptor["termAccessionNumber"],
                    ]

                    log(f"\t  - {', '.join(parts).strip(',').strip().strip(',')}")


if __name__ == "__main__":
    public_search(
        [
            "(chicken | turkey)",
            "--body",
            '{"assayTechniqueNameFilter": {"joinOperator": "and", "values": ["NMR"]}}',
        ]
    )



from typing import Dict, List, Union

from metabolights_utils.common import CamelCaseModel


class FtpLoginCredentials(CamelCaseModel):
    user_name: str = ""
    password: str = ""


class RestApiCredentials(CamelCaseModel):
    api_token: str = ""


class LoginCredentials(CamelCaseModel):
    ftp_login: Dict[str, FtpLoginCredentials] = {}
    rest_api_credentials: Dict[str, RestApiCredentials] = {}


class ResponseFileDescriptor(CamelCaseModel):
    created_at: Union[None, str] = None
    directory: bool = False
    file: Union[None, str] = None
    status: Union[None, str] = None
    timestamp: Union[None, str] = None
    type: Union[None, str] = None
    relative_path: Union[None, str] = None
    extension: Union[None, str] = None
    is_stop_folder: bool = False
    is_empty: bool = False


class StudyResponse(CamelCaseModel):
    study: List[ResponseFileDescriptor] = []
    upload_path: Union[None, str] = None
    obfuscation_code: Union[None, str] = None


class SubmittedStudySummary(CamelCaseModel):
    accession: Union[None, str] = None
    created_date: Union[None, str] = None
    release_date: Union[None, str] = None
    curation_request: Union[None, str] = None
    description: Union[None, str] = None
    status: Union[None, str] = None
    title: Union[None, str] = None
    updated: Union[None, str] = None


class SubmittedStudiesResponse(CamelCaseModel):
    data: List[SubmittedStudySummary] = []



import os
from typing import Callable, List, Union

import click

from metabolights_utils.commands.utils import split_to_lines
from metabolights_utils.common import sort_by_study_id
from metabolights_utils.models.enums import GenericMessageType
from metabolights_utils.provider import definitions
from metabolights_utils.provider.ftp_repository import MetabolightsFtpRepository


@click.command(name="list")
@click.option(
    "--local_path",
    "-p",
    default=definitions.default_local_repority_root_path,
    help="Local storage root path. Folder will be created if it does not exist.",
)
@click.option(
    "--ftp_server_url",
    "-f",
    default=f"{definitions.default_ftp_server_url}",
    help="FTP server URL where MetaboLights repository is hosted. ",
)
@click.option(
    "--ftp_root_directory",
    "-d",
    default=f"{definitions.default_remote_repository_root_directory}",
    help="MetaboLights study directory on FTP server URL. ",
)
@click.option(
    "--local_cache_path",
    "-c",
    default=definitions.default_local_repority_cache_path,
    help="Path to store cache files of FTP file indices, study models, etc.",
)
@click.option(
    "--use_only_local",
    "-l",
    is_flag=True,
    default=False,
    help="Use only current local directory without connecting FTP server.",
)
@click.argument("study_id", required=False)
@click.argument("subdirectory", required=False)
def public_list(
    study_id: Union[None, str] = None,
    subdirectory: Union[None, str] = None,
    local_path: Union[None, str] = None,
    local_cache_path: Union[None, str] = None,
    ftp_server_url: Union[None, str] = None,
    ftp_root_directory: Union[None, str] = None,
    use_only_local: bool = False,
):
    """
    List studies and study folder content. It works for both local and remote FTP repository.

    study_id (optional): MetaboLights study accession number (MTBLSxxxx). List all studies if not specified.

    subdirectory (optional): Subdirectory of study to list its content. List the study root folder if not specified..
    """
    client = MetabolightsFtpRepository(
        ftp_server_url=ftp_server_url,
        remote_repository_root_directory=ftp_root_directory,
        local_storage_root_path=local_path,
        local_storage_cache_path=local_cache_path,
    )
    if use_only_local:
        search = client.local_storage_root_path
        if study_id:
            study_id = study_id.upper()
            search = os.path.join(client.local_storage_root_path, study_id)
            if subdirectory:
                search = os.path.join(search, subdirectory.strip("/"))
        if search == client.local_storage_root_path:
            studies = [x for x in os.listdir(search) if x.startswith("MTBLS")]
            studies.sort(key=sort_by_study_id)
            if studies:
                if len(studies) > 500:
                    print_list("Local public study", studies, click.echo_via_pager)
                else:
                    print_list("Local public study", studies, click.echo)
            else:
                click.echo("No public studies found on local storage.")
        else:
            items = os.listdir(search)
            for item in items:
                click.echo(f"\t{item}")
        return

    if study_id and study_id.strip():
        study_id = study_id.upper().strip()
        content_response = client.list_study_directory(
            study_id=study_id, subdirectory=subdirectory
        )
        if not content_response.success:
            click.echo(f"FTP response error: {content_response.message}")
            exit(1)
        for item in content_response.files:
            click.echo(f"\t{item}")

        for item in content_response.folders:
            click.echo(f"\t{item}")
    else:
        studies, messages = client.list_studies()
        error = False
        if messages:
            for message in messages:
                if message.type in (
                    GenericMessageType.CRITICAL,
                    GenericMessageType.ERROR,
                ):
                    click.echo(f"{message.short}")
                    error = True
        if error:

            exit(1)

        if len(studies) > 500:
            print_list("FTP Public study", studies, click.echo_via_pager)
        else:
            print_list("FTP Public study", studies, click.echo)


def print_list(label: str, items: List[str], log: Callable):
    item_str = ", ".join(items)

    lines = split_to_lines(item_str, sep=", ", join_term="\t", max_line_size=100)
    result = f"{label} count: {len(items)}\n\t" + "\n\t".join([x for x in lines])
    log(result)


if __name__ == "__main__":
    public_list()



from typing import Union

import click
from jsonpath_ng import parse

from metabolights_utils.commands.utils import print_study_model_summary
from metabolights_utils.models.parser.enums import ParserMessageType
from metabolights_utils.provider import definitions
from metabolights_utils.provider.ftp_repository import MetabolightsFtpRepository


@click.command(no_args_is_help=True, name="describe")
@click.option(
    "--local_path",
    "-p",
    default=definitions.default_local_repority_root_path,
    help="Local storage root path. Folder will be created if it does not exist.",
)
@click.option(
    "--ftp_server_url",
    "-f",
    default=definitions.default_ftp_server_url,
    help="FTP server URL where MetaboLights repository is hosted. ",
)
@click.option(
    "--ftp_root_directory",
    "-d",
    default=definitions.default_remote_repository_root_directory,
    help="MetaboLights study directory on FTP server URL. ",
)
@click.option(
    "--use_only_local",
    "-l",
    is_flag=True,
    default=False,
    help="Use only current local directory without connecting FTP server.",
)
@click.option(
    "--override_local_files",
    "-o",
    is_flag=True,
    default=False,
    help="Downloads files and override current local copies. "
    "It is valid if there is no use_only_local option",
)
@click.option(
    "--load_folder_index",
    "-i",
    is_flag=True,
    default=False,
    help="Create a folder index to store file descriptors on FTP seerver. "
    "It is valid if there is no use_only_local option.",
)
@click.option(
    "--use_study_model_cache",
    is_flag=True,
    default=True,
    help="Use only current local directory without connecting FTP server.",
)
@click.option(
    "--local_cache_path",
    "-c",
    default=definitions.default_local_repority_cache_path,
    help="Path to store cache files of FTP file indices, study models, etc.",
)
@click.argument("study_id")
@click.argument("jsonpath", required=False)
def public_describe(
    study_id: str = "",
    jsonpath: str = "",
    use_only_local: bool = False,
    local_path: Union[None, str] = None,
    local_cache_path: Union[None, str] = None,
    ftp_server_url: Union[None, str] = None,
    ftp_root_directory: Union[None, str] = None,
    override_local_files: bool = False,
    load_folder_index: bool = False,
    use_study_model_cache: bool = True,
):
    """
    View summary of any public study content. Run jsonpath expression to filter MetaboLights study model.

    study_id: MetaboLights study accession number (MTBLSxxxx).

    jsonpath (optional): jsonpath expression to filter study model.

    Print summary of study model if not specified.

        Example jsonpath expressions: "$.investigation.studies[0].title", "$.investigation.studies[0].study_protocols.protocols[*].name", "$.assays[*].*.table.columns[*]" - print column names

        Note: jsonpath expressions should be quoted with double quotes.
    """
    study_id = study_id.upper().strip()
    client = MetabolightsFtpRepository(
        ftp_server_url=ftp_server_url,
        remote_repository_root_directory=ftp_root_directory,
        local_storage_root_path=local_path,
        local_storage_cache_path=local_cache_path,
    )

    local_path = client.local_storage_root_path

    model, messages = client.load_study_model(
        study_id=study_id,
        local_path=local_path,
        use_only_local_path=use_only_local,
        override_local_files=override_local_files,
        load_folder_metadata=load_folder_index,
        use_study_model_cache=use_study_model_cache,
    )

    if (
        not model
        or not model.investigation
        or not model.investigation.studies
        or not model.investigation.studies[0]
    ):
        click.echo("Failure: There are parser error or invalid model.")
        exit(1)

    if jsonpath:
        try:
            json_data = model.model_dump()

            jsonpath_expr = parse(jsonpath)
            match_values = [match.value for match in jsonpath_expr.find(json_data)]
            if match_values:
                click.echo(
                    click.style(f"{study_id}: '{jsonpath}' search result:", fg="green")
                )
                if len(match_values) == 1:
                    click.echo(match_values[0])
                else:
                    for match_item in match_values:
                        click.echo(f"{match_item}")
            else:
                click.echo(click.style(f"{study_id}: '{jsonpath}' no match", fg="red"))
        except Exception as ex:
            click.echo(f"jsonpath '{jsonpath}' expression error: {str(ex)}")
            exit(1)
    else:
        if model.parser_messages:
            for file, messages in model.parser_messages.items():
                for message in messages:
                    if message.type in (
                        ParserMessageType.CRITICAL,
                        ParserMessageType.ERROR,
                    ):
                        click.echo(f"{study_id} {file}: {message.short}")
        click.echo(click.style("MetaboLights study model summary.", fg="green"))
        print_study_model_summary(model, log=click.echo)


if __name__ == "__main__":
    public_describe(["MTBLS1"])



import pathlib
import uuid
from typing import List

from metabolights_utils.tsv import model as actions
from metabolights_utils.tsv.actions.base import BaseTsvAction


class DeleteRowsTsvAction(BaseTsvAction):
    def apply_action(
        self,
        source_file_path: pathlib.Path,
        target_file_path: pathlib.Path,
        action: actions.TsvDeleteRowsAction,
        read_encoding: str = "utf-8",
        write_encoding: str = "utf-8",
    ) -> actions.TsvActionResult:
        result: actions.TsvActionResult = actions.TsvActionResult(action=action)
        if action.action_type != actions.TsvActionType.DELETE_ROW:
            result.message = "Action name is not valid"
            return result

        action: actions.TsvDeleteRowsAction = action
        target_row_indices: List[int] = action.current_row_indices

        if not target_row_indices:
            result.message = "There is not row index"
            return result

        row_indices = target_row_indices.copy()
        row_indices.sort()

        if not action.id:
            uuid_value = str(uuid.uuid4().hex)
            action.id = uuid_value

        try:
            with open(source_file_path, "r", encoding=read_encoding) as source:
                header_line = source.readline()
                header_names = header_line.strip("\n").split("\t")

                with open(target_file_path, "w", encoding=write_encoding) as target:
                    self.write_row(target, header_names)
                    row_index = -1
                    for line in source:
                        row_index += 1
                        if row_index in row_indices:
                            row_indices.remove(row_index)
                            continue
                        else:
                            target.write(line)

                    if len(row_indices):
                        result.message = f"Invalid row indices {', '.join(row_indices)}"
                        return result
            result.success = True
        except Exception as exc:
            result.message = f"{str(exc)}"

        return result



import os

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class CamelCaseModel(BaseModel):
    """Base model class to convert python attributes to camel case"""

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        json_schema_serialization_defaults_required=True,
    )


def sort_by_study_id(key: str):
    if key:
        val = os.path.basename(key).upper().replace("MTBLS", "")
        if val.isnumeric():
            value = int(val)
            return value if value > 0 else -1
    return -1



import click

from metabolights_utils import __VERSION__
from metabolights_utils.commands.model.model_cli import model_cli
from metabolights_utils.commands.public.public_cli import public_cli
from metabolights_utils.commands.submission.submission_cli import submission_cli


@click.group()
@click.version_option(__VERSION__)
def cli():
    pass


cli.add_command(public_cli)
cli.add_command(submission_cli)
cli.add_command(model_cli)



from typing import List

from metabolights_utils.isatab.default.base_isa_table_file import BaseIsaTableFileReader


class DefaultSampleFileReader(BaseIsaTableFileReader):
    patterns = [
        [r"^[ ]*(Source Name)[ ]*$", ""],
        [r"^[ ]*Characteristics[ ]*\[[ ]*(\w[ -~]*)[ ]*\](.\d+)?$", "Characteristics"],
        [r"^[ ]*(Protocol REF)[ ]*(.\d+)?[ ]*$", "Protocol"],
        [r"^[ ]*(Sample Name)[ ]*$", ""],
        [r"^[ ]*Factor[ ]+Value[ ]*\[[ ]*(\w[ -~]*)[ ]*\](.\d+)?$", "Factor Value"],
        [r"^[ ]*Comment[ ]*\[[ ]*(\w[ -~]*)[ ]*\](.\d+)?$", "Comment"],
    ]

    def __init__(self, results_per_page=100) -> None:
        super().__init__(results_per_page=results_per_page)

    def get_expected_patterns(self) -> List[List[str]]:
        return DefaultSampleFileReader.patterns



import logging
import os
from typing import Any, Union

from metabolights_utils.isatab import Writer
from metabolights_utils.models.isa.common import IsaTableFile
from metabolights_utils.models.metabolights.model import MetabolightsStudyModel

logger = logging.getLogger(__name__)


class IsaFileUtils:

    @staticmethod
    async def save_metabolights_study_model(
        mtbls_model: MetabolightsStudyModel,
        output_dir: str,
        values_in_quotation_mark: bool = False,
        investigation_module_name: Union[None, str] = None,
    ):
        if not os.path.exists(output_dir):
            logger.info(f"Study save folder % is created", output_dir)
            os.makedirs(output_dir, exist_ok=True)
        i_file_name = mtbls_model.investigation_file_path
        logger.info("Saving investigation file: %s", i_file_name)
        Writer.get_investigation_file_writer().write(
            mtbls_model.investigation,
            f"{output_dir}/{mtbls_model.investigation_file_path}",
            values_in_quotation_mark=values_in_quotation_mark,
            investigation_module_name=investigation_module_name,
        )
        for isa_table_files in (
            mtbls_model.samples,
            mtbls_model.assays,
            mtbls_model.metabolite_assignments,
        ):
            for isa_table_file in isa_table_files.values():
                i_file_name = isa_table_file.file_path
                await IsaFileUtils.save_isa_table(
                    isa_table_file,
                    f"{output_dir}/{isa_table_file.file_path}",
                    values_in_quotation_mark=values_in_quotation_mark,
                )

    @staticmethod
    def sanitise_data(self, value: Union[None, Any]):
        if isinstance(value, list):
            for idx, val in enumerate(value):
                value[idx] = IsaFileUtils.sanitise_single_value(val)
            return value
        return IsaFileUtils.sanitise_single_value(value)

    @staticmethod
    def sanitise_single_value(self, value: Union[None, Any]):
        if value is None:
            return ""
        return (
            str(value).replace("\n", " ").replace("\r", " ").replace("\t", " ").strip()
        )

    @staticmethod
    async def save_isa_table(
        isa_table_file: IsaTableFile,
        file_path: str,
        values_in_quotation_mark=False,
    ):
        column_order_map = {}
        column_header_map = {}
        data = isa_table_file.table.data
        for column_model in isa_table_file.table.headers:
            column_order_map[column_model.column_index] = column_model.column_name
            column_header_map[column_model.column_index] = column_model.column_header
        parent_dir = os.path.dirname(file_path)
        basename = os.path.dirname(file_path)
        if not os.path.exists(parent_dir):
            logger.info(f"Study save folder % is created", parent_dir)
            os.makedirs(parent_dir, exist_ok=True)

        logger.info("Saving file: %s", basename)
        with open(file_path, "w") as f:
            if values_in_quotation_mark:
                header = [
                    f'"{column_header_map[idx]}"'
                    for idx in range(len(column_header_map))
                ]
            else:
                header = [
                    column_header_map[idx].strip('"')
                    for idx in range(len(column_header_map))
                ]
            f.write("\t".join(header) + "\n")

            column_names = [
                column_order_map[idx] for idx in range(len(column_order_map))
            ]
            for row_idx in range(len(data[column_names[0]])):
                row = [data[column_name][row_idx] for column_name in column_names]
                for idx, cell in enumerate(row):
                    cell = IsaFileUtils.sanitise_data(cell) if row[idx] else ""
                    if values_in_quotation_mark:
                        cell = f'"{cell}"'
                    else:
                        cell = cell.strip('"')
                    row[idx] = cell
                f.write("\t".join(row) + "\n")



import json
import os
from typing import Dict, Union
from metabolights_utils.models.metabolights import metabolights_study, model


def get_study_model_schema(schema_path: Union[str, None] = None) -> Dict:
    if not schema_path:
        schema_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "study_model_schema_v1.0.json",
        )
    with open(schema_path, "r", encoding="utf-8") as sf:
        study_model_jsonschema = json.load(sf)
    return study_model_jsonschema


__all__ = ["metabolights_study", "model", "get_study_model_schema"]



import os
import pathlib
import shutil
from abc import ABC, abstractmethod
from io import IOBase
from typing import List

from metabolights_utils.tsv import model as actions


class TsvActionException(Exception):
    def __init__(self, message: str = "") -> None:
        self.message = message


class BaseTsvAction(ABC):
    @abstractmethod
    def apply_action(
        self,
        source_file_path: pathlib.Path,
        target_file_path: pathlib.Path,
        action: actions.TsvAction,
        read_encoding: str = "utf-8",
        write_encoding: str = "utf-8",
    ) -> actions.TsvActionResult:
        pass

    def delete_file(self, path):
        if os.path.exists(path):
            if os.path.islink(path):
                os.unlink(path)
            elif os.path.isfile(path):
                os.remove(path)
            elif os.path.isdir(path):
                shutil.rmtree(path)

    def get_updated_row(self, empty_row, input_row: actions.TsvRowData):
        column_indices = range(len(empty_row))
        data_values = input_row.values

        def merge_method(x):
            return data_values[x] if data_values and x in data_values else empty_row[x]

        map_result = map(merge_method, column_indices)
        return list(map_result)

    def write_row(self, file_buffer: IOBase, row: List[str]):
        new_row_string = "\t".join(row) + "\n"
        file_buffer.write(new_row_string)



from metabolights_utils.provider import (
    ftp,
    async_provider,
    definitions,
    ftp_repository,
    local_folder_metadata_collector,
    study_provider,
    submission_model,
    submission_repository,
    utils,
)

__all__ = [
    "ftp",
    "async_provider",
    "definitions",
    "ftp_repository",
    "local_folder_metadata_collector",
    "study_provider",
    "submission_model",
    "submission_repository",
    "utils",
]



import logging
import os
import re
from typing import List, Tuple, Union

import unidecode

logger = logging.getLogger(__name__)


class MetabolightsFileNameUtils(object):
    @staticmethod
    def sanitise_filename(
        filename: str,
        replacement_chars_pattern: Union[None, str] = None,
        allow_spaces: bool = False,
    ) -> str:
        if not filename or not filename.strip():
            return ""

        if not replacement_chars_pattern:
            replacement_chars_pattern = "[^/a-zA-Z0-9_.-]"
        new_filename = unidecode.unidecode(filename.strip())
        new_filename = new_filename.replace("+", "_PLUS_")
        if not allow_spaces:
            new_filename = "_".join(
                [x.strip() for x in new_filename.strip().split() if x.strip()]
            )

        new_filename = re.sub(replacement_chars_pattern, "_", new_filename)
        if new_filename != filename:
            logger.debug("'%s' is converted to '%s'", filename, new_filename)
        return new_filename


def join_path(*args: Tuple[str]):
    target_sep = "/" if os.sep == "/" else "\\"
    source_sep = "/" if os.sep == "\\" else "/"
    inputs = list([x for x in args if x and x.strip()])
    _path = target_sep.join(inputs).replace(source_sep, target_sep)
    return _path



import io
import json
import logging
import os
import re
import zipfile
from typing import Any, Dict, List, Set, Tuple, Union

import httpx

from metabolights_utils.models.isa.assay_file import AssayFile
from metabolights_utils.models.isa.common import AssayTechnique
from metabolights_utils.models.metabolights.model import MetabolightsStudyModel

logger = logging.getLogger(__name__)


def is_metadata_file(file_path: str):
    if os.path.exists(file_path) and os.path.isfile(file_path):
        filename = os.path.basename(file_path)
        return is_metadata_filename_pattern(filename)
    return False


def is_metadata_filename_pattern(filename: str):
    if not filename:
        return False
    if len(filename) > 6:
        if filename[:2] in ("a_", "s_", "i_") and filename.endswith(".txt"):
            return True
        elif filename.startswith("m_") and filename.endswith(".tsv"):
            return True
    return False


def download_file_from_rest_api(
    url: str,
    local_file_path: str,
    timeout: Union[None, int] = None,
    headers: Union[None, Dict[str, Any]] = None,
    parameters: Union[None, Dict[str, Any]] = None,
    modification_time: Union[None, int, float] = None,
    is_zip_response: bool = False,
) -> Tuple[bool, str]:
    try:
        directory = os.path.dirname(local_file_path)
        os.makedirs(directory, exist_ok=True)
        data_bytes = io.BytesIO()

        with httpx.stream(
            "GET",
            url,
            timeout=timeout,
            headers=headers,
            params=parameters,
        ) as response:
            response.raise_for_status()

            for data in response.iter_bytes():
                data_bytes.write(data)

            if is_zip_response:
                with zipfile.ZipFile(data_bytes) as zip_file:
                    zip_file.extractall(directory)
            else:
                with open(local_file_path, "wb") as f:
                    f.write(data_bytes.read())

            if modification_time:
                if is_zip_response:
                    for file in os.listdir(directory):
                        file_path = os.path.join(directory, file)
                        os.utime(file_path, (modification_time, modification_time))
                else:
                    os.utime(local_file_path, (modification_time, modification_time))
            return True, None
    except Exception as ex:
        logger.exception(str(ex))
        return False, str(ex)


def rest_api_get(
    url: str,
    timeout: Union[None, int] = None,
    headers: Union[None, Dict[str, Any]] = None,
    parameters: Union[None, Dict[str, Any]] = None,
):
    try:
        response = httpx.get(
            url=url,
            timeout=timeout,
            headers=headers,
            params=parameters,
        )
        if response and response.status_code in (200, 201):
            data = json.loads(response.text)
            return data, None
        return None, response.text
    except Exception as ex:
        return None, str(ex)


def rest_api_post(
    url: str,
    timeout: Union[None, int] = None,
    headers: Union[None, Dict[str, Any]] = None,
    parameters: Union[None, Dict[str, Any]] = None,
    json_body: Union[None, Dict[str, Any]] = None,
):
    try:
        response = httpx.post(
            url=url, timeout=timeout, headers=headers, params=parameters, json=json_body
        )
        if response and response.status_code in (200, 201):
            data = json.loads(response.text)
            return data, None
        return None, response.text
    except Exception as ex:
        return None, str(ex)


def get_unique_file_extensions(files: Set[str]) -> Set[str]:
    extensions = set()

    for item in files:
        name1, ext1 = os.path.splitext(item)
        _, ext2 = os.path.splitext(name1)
        if ext1:
            if ext2 and len(ext2) < 6:
                extensions.add(f"{ext2.lower()}{ext1.lower()}")
            else:
                extensions.add(f"{ext1.lower()}")
    return extensions


def find_assay_technique(
    model: MetabolightsStudyModel,
    assay_file: AssayFile,
    assay_file_subset: AssayFile,
):
    nmr_assay_name = "NMR Assay Name"
    ms_assay_name = "MS Assay Name"
    # if nmr_assay_name and ms_assay_name are in the assay file, it is not well-structured assay
    # return empty
    if (
        nmr_assay_name in assay_file.table.columns
        and ms_assay_name in assay_file.table.columns
    ):
        return AssayTechnique()

    if nmr_assay_name in assay_file.table.columns:
        columns = get_assay_technique_search_columns("MRImaging")
        columns_in_assay_file = {x for x in assay_file.table.columns if x in columns}
        if len(columns) > 0 and len(columns_in_assay_file) == len(columns):
            return assay_techniques["MRImaging"]
    elif ms_assay_name in assay_file.table.columns:
        # search unique column names for each assay technique
        # return technique if assay has the relevant unique column names
        technique_names = [
            "DI-MS",
            "FIA-MS",
            "CE-MS",
            "SPE-IMS-MS",
            "TD-GC-MS",
            "LC-DAD",
            "MSImaging",
        ]
        for technique_name in technique_names:
            columns = get_assay_technique_search_columns(technique_name)
            columns_in_assay_file = {
                x for x in assay_file.table.columns if x in columns
            }
            if len(columns) > 0 and len(columns_in_assay_file) == len(columns):
                return assay_techniques[technique_name]
        # LC-MS or GC-MS difference
        column_type = "Parameter Value[Column type]"
        if column_type in assay_file_subset.table.columns:
            if assay_file_subset.table.data[column_type]:
                values = assay_file_subset.table.data[column_type]
                for i in range(len(values) if len(values) < 5 else 5):
                    if values[i]:
                        column_type = values[i].lower()
                        if "hilic" in column_type or "reverse" in column_type:
                            return assay_techniques["LC-MS"]
                        elif (
                            "low polarity" in column_type
                            or "high polarity" in column_type
                            or "medium polarity" in column_type
                        ):
                            return assay_techniques["GC-MS"]
    investigation_assay = None
    # check i_Investigation.txt file
    # Find the related assay and check technology platform
    # if technology platform contains assay technique name return the relavant assay technique
    for study in model.investigation.studies:
        for assay in study.study_assays.assays:
            if assay.file_name == assay_file.file_path:
                investigation_assay = assay
                for technique in assay_technique_keywords:
                    if technique in assay.technology_platform:
                        return assay_techniques[assay_technique_keywords[technique]]

    # try database curator tags to classify the assay
    if model.study_db_metadata and model.study_db_metadata.study_types:
        study_type_str = ",".join(model.study_db_metadata.study_types)
        if study_type_str:
            for pattern, technique_name in manual_assignments_patterns.items():
                result = re.search(pattern, study_type_str)
                if result:
                    return assay_techniques[technique_name]
    # if there is no match, use assay technology type and return generic assay technique
    if (
        investigation_assay
        and investigation_assay.technology_type
        and investigation_assay.technology_type.term
    ):
        if "mass spectrometry" in investigation_assay.technology_type.term:
            return assay_techniques["MS"]
        elif "NMR spectrometry" in investigation_assay.technology_type.term:
            return assay_techniques["NMR"]

    if nmr_assay_name in assay_file.table.columns:
        return assay_techniques["NMR"]
    return AssayTechnique()


assay_techniques = {
    "LC-MS": AssayTechnique(
        name="LC-MS",
        mainTechnique="MS",
        technique="LC-MS",
        subTechnique="LC",
    ),
    "LC-DAD": AssayTechnique(
        name="LC-DAD",
        mainTechnique="MS",
        technique="LC-MS",
        subTechnique="LC-DAD",
    ),
    "GC-MS": AssayTechnique(
        name="GC-MS",
        mainTechnique="MS",
        technique="GC-MS",
        subTechnique="GC",
    ),
    "GCxGC-MS": AssayTechnique(
        name="GCxGC-MS",
        mainTechnique="MS",
        technique="GC-MS",
        subTechnique="Tandem (GCxGC)",
    ),
    "GC-FID": AssayTechnique(
        name="GC-FID",
        mainTechnique="MS",
        technique="GC-MS",
        subTechnique="GC-FID",
    ),
    "MS": AssayTechnique(
        name="MS",
        mainTechnique="MS",
        technique="Direct Injection",
        subTechnique="MS",
    ),
    "DI-MS": AssayTechnique(
        name="DI-MS",
        mainTechnique="MS",
        technique="Direct Injection",
        subTechnique="Direct infusion (DI)",
    ),
    "FIA-MS": AssayTechnique(
        name="FIA-MS",
        mainTechnique="MS",
        technique="Direct Injection",
        subTechnique="Flow injection analysis (FIA)",
    ),
    "CE-MS": AssayTechnique(
        name="CE-MS",
        mainTechnique="MS",
        technique="Direct Injection",
        subTechnique="Capillary electrophoresis (CE)",
    ),
    "MALDI-MS": AssayTechnique(
        name="MALDI-MS",
        mainTechnique="MS",
        technique="Direct Injection",
        subTechnique="Matrix-assisted laser desorption-ionisation imaging mass spectrometry (MALDI)",
    ),
    "SPE-IMS-MS": AssayTechnique(
        name="SPE-IMS-MS",
        mainTechnique="MS",
        technique="Direct Injection",
        subTechnique="Solid-Phase Extraction Ion Mobility Spectrometry (SPE-IMS)",
    ),
    "MSImaging": AssayTechnique(
        name="MSImaging",
        mainTechnique="MS",
        technique="MS Imaging",
        subTechnique="MS Imaging",
    ),
    "NMR": AssayTechnique(
        name="NMR",
        mainTechnique="NMR",
        technique="NMR",
        subTechnique="Nuclear magnetic resonance",
    ),
    "MRImaging": AssayTechnique(
        name="MRImaging",
        mainTechnique="NMR",
        technique="MRI",
        subTechnique="Magnetic resonance imaging",
    ),
}


assay_technique_keywords: Dict[str, str] = {
    "Liquid Chromatography MS": "LC-MS",
    "Diode array detection MS": "LC-DAD",
    "Gas Chromatography MS": "GC-MS",
    "Tandem Gas Chromatography MS": "GCxGC-MS",
    "Flame ionisation detector MS": "GC-FID",
    "Mass spectrometry": "MS",
    "Direct infusion MS": "DI-MS",
    "Flow injection analysis MS": "FIA-MS",
    "Capillary electrophoresis MS": "CE-MS",
    "Matrix-assisted laser desorption-ionisation imaging MS": "MALDI-MS",
    "Solid-Phase Extraction Ion Mobility Spectrometry MS": "SPE-IMS-MS",
    "MS Imaging": "MSImaging",
    "Nuclear Magnetic Resonance (NMR)": "NMR",
    "Magnetic resonance imaging": "MRImaging",
}

assay_technique_labels: Dict[str, str] = {
    "LC-MS": "Liquid Chromatography MS",
    "LC-DAD": "Diode array detection MS",
    "GC-MS": "Gas Chromatography MS",
    "GCxGC-MS": "Tandem Gas Chromatography MS",
    "GC-FID": "Flame ionisation detector MS",
    "MS": "Mass spectrometry",
    "DI-MS": "Direct infusion MS",
    "FIA-MS": "Flow injection analysis MS",
    "CE-MS": "Capillary electrophoresis MS",
    "MALDI-MS": "Matrix-assisted laser desorption-ionisation imaging MS",
    "SPE-IMS-MS": "Solid-Phase Extraction Ion Mobility Spectrometry MS",
    "MSImaging": "MS Imaging",
    "NMR": "Nuclear Magnetic Resonance (NMR)",
    "MRImaging": "Magnetic resonance imaging",
}

manual_assignments_patterns = {
    r".*imaging.*MS.*": "MSImaging",
    r".*MS.*Imaging.*": "MSImaging",
    r".*CE-.*-MS.*": "CE-MS",
    r".*DI-.*-MS.*": "DI-MS",
    r".*FIA-.*-MS.*": "FIA-MS",
    r".*GC-.*-MS.*": "GC-MS",
    r".*LC-.*-MS.*": "LC-MS",
    r".*MALDI-.*-MS.*": "MALDI-MS",
}

assay_technique_non_protocol_search_columns = {
    "TD-GC-MS": ["Thermal Desorption unit", "Cryogenic Trap"],
    "MRImaging": [
        "Parameter Value[Tomography]",
        "Parameter Value[Spatial resolution]",
    ],
    "DI-MS": [
        "Parameter Value[DI Instrument]",
    ],
    "FIA-MS": [
        "Parameter Value[FIA Instrument]",
    ],
    "CE-MS": [
        "Parameter Value[CE Instrument]",
    ],
    "GCxGC-MS": [
        "Parameter Value[Column model 1]",
        "Parameter Value[Column model 2]",
        "Parameter Value[Column type 1]" "Parameter Value[Column type 2]",
    ],
    "SPE-IMS-MS": [
        "Parameter Value[SPE-IMS Instrument",
    ],
    "MSImaging": [
        "Parameter Value[Sectioning instrument]",
        "Parameter Value[Instrument manufacturer]",
    ],
    "LC-DAD": [
        "Parameter Value[Signal range]",
        "Parameter Value[Resolution]",
        "Parameter Value[Detector]",
    ],
}


def get_assay_technique_search_columns(technique_name: str) -> List[str]:
    if technique_name in assay_technique_non_protocol_search_columns:
        return assay_technique_non_protocol_search_columns[technique_name]
    return []


default_assay_protocols = {
    "CE-MS": {
        "name": "CE-MS",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {"name": "Extraction", "parameters": ["Post Extraction", "Derivatization"]},
            {
                "name": "Capillary Electrophoresis",
                "parameters": ["CE Instrument", "Column model", "Column type"],
            },
            {
                "name": "Mass spectrometry",
                "parameters": [
                    "Scan polarity",
                    "Scan m/z range",
                    "Instrument",
                    "Ion source",
                    "Mass analyzer",
                ],
            },
            {"name": "Data transformation", "parameters": []},
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": "Capillary electrophoresis MS",
        "technology_type": "mass spectrometry",
    },
    "DI-MS": {
        "name": "DI-MS",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {"name": "Extraction", "parameters": ["Post Extraction", "Derivatization"]},
            {"name": "Direct infusion", "parameters": ["DI Instrument"]},
            {
                "name": "Mass spectrometry",
                "parameters": [
                    "Scan polarity",
                    "Scan m/z range",
                    "Instrument",
                    "Ion source",
                    "Mass analyzer",
                ],
            },
            {"name": "Data transformation", "parameters": []},
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": "Direct infusion MS",
        "technology_type": "mass spectrometry",
    },
    "FIA-MS": {
        "name": "FIA-MS",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {"name": "Extraction", "parameters": ["Post Extraction", "Derivatization"]},
            {"name": "Flow Injection Analysis", "parameters": ["FIA Instrument"]},
            {
                "name": "Mass spectrometry",
                "parameters": [
                    "Scan polarity",
                    "Scan m/z range",
                    "Instrument",
                    "Ion source",
                    "Mass analyzer",
                ],
            },
            {"name": "Data transformation", "parameters": []},
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": "Flow injection analysis MS",
        "technology_type": "mass spectrometry",
    },
    "GC-FID": {
        "name": "GC-FID",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {"name": "Extraction", "parameters": ["Post Extraction", "Derivatization"]},
            {
                "name": "Chromatography",
                "parameters": [
                    "Chromatography Instrument",
                    "Autosampler model",
                    "Column model",
                    "Column type",
                    "Guard column",
                    "Detector",
                    "Temperature",
                ],
            },
            {"name": "Data transformation", "parameters": []},
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": "Flame ionisation detector MS",
        "technology_type": "mass spectrometry",
    },
    "GC-MS": {
        "name": "GC-MS",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {"name": "Extraction", "parameters": ["Post Extraction", "Derivatization"]},
            {
                "name": "Chromatography",
                "parameters": [
                    "Chromatography Instrument",
                    "Autosampler model",
                    "Column model",
                    "Column type",
                    "Guard column",
                ],
            },
            {
                "name": "Mass spectrometry",
                "parameters": [
                    "Scan polarity",
                    "Scan m/z range",
                    "Instrument",
                    "Ion source",
                    "Mass analyzer",
                ],
            },
            {"name": "Data transformation", "parameters": []},
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": "Gas Chromatography MS",
        "technology_type": "mass spectrometry",
    },
    "GCxGC-MS": {
        "name": "GCxGC-MS",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {"name": "Extraction", "parameters": ["Post Extraction", "Derivatization"]},
            {
                "name": "Chromatography",
                "parameters": [
                    "Chromatography Instrument",
                    "Autosampler model",
                    "Column model 1",
                    "Column type 1",
                    "Guard column",
                    "Column model 2",
                    "Column type 2",
                ],
            },
            {
                "name": "Mass spectrometry",
                "parameters": [
                    "Scan polarity",
                    "Scan m/z range",
                    "Instrument",
                    "Ion source",
                    "Mass analyzer",
                ],
            },
            {"name": "Data transformation", "parameters": []},
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": "Tandem Gas Chromatography MS",
        "technology_type": "mass spectrometry",
    },
    "LC-DAD": {
        "name": "LC-DAD",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {"name": "Extraction", "parameters": ["Post Extraction", "Derivatization"]},
            {
                "name": "Chromatography",
                "parameters": [
                    "Chromatography Instrument",
                    "Autosampler model",
                    "Column model",
                    "Column type",
                    "Guard column",
                    "Detector",
                    "Signal range",
                    "Resolution",
                ],
            },
            {"name": "Data transformation", "parameters": []},
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": "Diode array detection MS",
        "technology_type": "mass spectrometry",
    },
    "LC-MS": {
        "name": "LC-MS",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {"name": "Extraction", "parameters": ["Post Extraction", "Derivatization"]},
            {
                "name": "Chromatography",
                "parameters": [
                    "Chromatography Instrument",
                    "Autosampler model",
                    "Column model",
                    "Column type",
                    "Guard column",
                ],
            },
            {
                "name": "Mass spectrometry",
                "parameters": [
                    "Scan polarity",
                    "Scan m/z range",
                    "Instrument",
                    "Ion source",
                    "Mass analyzer",
                ],
            },
            {"name": "Data transformation", "parameters": []},
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": "Liquid Chromatography MS",
        "technology_type": "mass spectrometry",
    },
    "MALDI-MS": {
        "name": "MALDI-MS",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {"name": "Extraction", "parameters": ["Post Extraction", "Derivatization"]},
            {
                "name": "Mass spectrometry",
                "parameters": [
                    "Scan polarity",
                    "Scan m/z range",
                    "Instrument",
                    "Ion source",
                    "Mass analyzer",
                ],
            },
            {"name": "Data transformation", "parameters": []},
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": [
            "Matrix-assisted laser desorption-ionisation imaging MS"
        ],
        "technology_type": "mass spectrometry",
    },
    "MRImaging": {
        "name": "MRImaging",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {
                "name": "Magnetic resonance imaging",
                "parameters": ["Instrument", "Tomography", "Temperature"],
            },
            {
                "name": "In vivo magnetic resonance spectroscopy",
                "parameters": [
                    "Spatial resolution",
                    "Field of view",
                    "Matrix",
                    "Magnetic pulse sequence name",
                    "Voxel size",
                    "Localisation pulse sequence name",
                    "Number of transients",
                    "Water inhibition pulse sequence name",
                    "Magnetic field strength",
                ],
            },
            {
                "name": "In vivo magnetic resonance assay",
                "parameters": ["NMR tube type", "Solvent", "Sample pH", "Temperature"],
            },
            {"name": "Data transformation", "parameters": []},
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": "Magnetic resonance imaging",
        "technology_type": ["NMR spectroscopy"],
    },
    "MS": {
        "name": "MS",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {"name": "Extraction", "parameters": ["Post Extraction", "Derivatization"]},
            {
                "name": "Mass spectrometry",
                "parameters": [
                    "Scan polarity",
                    "Scan m/z range",
                    "Instrument",
                    "Ion source",
                    "Mass analyzer",
                ],
            },
            {"name": "Data transformation", "parameters": []},
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": "Mass spectrometry",
        "technology_type": "mass spectrometry",
    },
    "MSImaging": {
        "name": "MSImaging",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {
                "name": "Preparation",
                "parameters": [
                    "Sample mounting",
                    "Sample preservation",
                    "Tissue modification",
                    "Sectioning instrument",
                    "Section thickness",
                    "Matrix",
                    "Matrix application",
                ],
            },
            {
                "name": "Mass spectrometry",
                "parameters": [
                    "Scan polarity",
                    "Scan m/z range",
                    "Instrument",
                    "Instrument manufacturer",
                    "Ion source",
                    "Mass analyzer",
                    "Solvent",
                    "Target material",
                    "Spatial resolution",
                    "Pixel size x",
                    "Pixel size y",
                    "Max count of pixel x",
                    "Max count of pixel y",
                    "Max dimension x",
                    "Max dimension y",
                    "Inlet type",
                    "Detector",
                    "Detector mode",
                    "Resolving power",
                    "Resolving power m/z",
                    "Native spectrum identifier format",
                    "Data file content",
                    "Spectrum representation",
                    "Raw data file format",
                    "Instrument software",
                    "Instrument software version",
                    "Line scan direction",
                    "Linescan sequence",
                    "Scan pattern",
                    "Scan type",
                    "Number of scans",
                ],
            },
            {
                "name": "Histology",
                "parameters": ["Stain", "High-res image", "Low-res image"],
            },
            {
                "name": "Data transformation",
                "parameters": [
                    "Data Transformation software",
                    "Data Transformation software version",
                ],
            },
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": "MS Imaging",
        "technology_type": "mass spectrometry",
    },
    "NMR": {
        "name": "NMR",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {"name": "Extraction", "parameters": ["Extraction Method"]},
            {
                "name": "NMR sample",
                "parameters": ["NMR tube type", "Solvent", "Sample pH", "Temperature"],
            },
            {
                "name": "NMR spectroscopy",
                "parameters": [
                    "Instrument",
                    "NMR Probe",
                    "Number of transients",
                    "Pulse sequence name",
                    "Magnetic field strength",
                ],
            },
            {"name": "NMR assay", "parameters": []},
            {"name": "Data transformation", "parameters": []},
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": "Nuclear Magnetic Resonance (NMR)",
        "technology_type": ["NMR spectroscopy"],
    },
    "SPE-IMS-MS": {
        "name": "SPE-IMS-MS",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {"name": "Extraction", "parameters": ["Post Extraction", "Derivatization"]},
            {
                "name": "Solid-Phase Extraction Ion Mobility Spectrometry",
                "parameters": ["SPE-IMS Instrument", "Cartridge type"],
            },
            {
                "name": "Mass spectrometry",
                "parameters": [
                    "Scan polarity",
                    "Scan m/z range",
                    "Instrument",
                    "Ion source",
                    "Mass analyzer",
                ],
            },
            {"name": "Data transformation", "parameters": []},
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": ["Solid-Phase Extraction Ion Mobility Spectrometry MS"],
        "technology_type": "mass spectrometry",
    },
    "TD-GC-MS": {
        "name": "TD-GC-MS",
        "protocols": [
            {"name": "Sample collection", "parameters": []},
            {"name": "Extraction", "parameters": ["Post Extraction", "Derivatization"]},
            {
                "name": "Chromatography",
                "parameters": [
                    "Chromatography Instrument",
                    "Autosampler model",
                    "Column model",
                    "Column type",
                ],
            },
            {
                "name": "Mass spectrometry",
                "parameters": [
                    "Scan polarity",
                    "Scan m/z range",
                    "Instrument",
                    "Ion source",
                    "Mass analyzer",
                ],
            },
            {"name": "Data transformation", "parameters": []},
            {"name": "Metabolite identification", "parameters": []},
        ],
        "technology_platform": "Thermal Desorption-Gas Chromatography MS",
        "technology_type": "mass spectrometry",
    },
}



import pathlib
import uuid
from typing import Dict

from metabolights_utils.tsv import model as actions
from metabolights_utils.tsv.actions.base import BaseTsvAction


class UpdateColumnHeadersTsvAction(BaseTsvAction):
    def apply_action(
        self,
        source_file_path: pathlib.Path,
        target_file_path: pathlib.Path,
        action: actions.TsvUpdateColumnHeaderAction,
        read_encoding: str = "utf-8",
        write_encoding: str = "utf-8",
    ) -> actions.TsvActionResult:
        result: actions.TsvActionResult = actions.TsvActionResult(action=action)
        if action.action_type != actions.TsvActionType.UPDATE_COLUMN_HEADER:
            result.message = "Action name is not valid"
            return result

        action: actions.TsvUpdateColumnHeaderAction = action

        headers: Dict[int, str] = action.new_headers if action.new_headers else {}

        column_indices = list(headers.keys())
        column_indices.sort()

        if not action.id:
            uuid_value = str(uuid.uuid4().hex)
            action.id = uuid_value

        try:
            with open(source_file_path, "r", encoding=read_encoding) as source:
                header_line = source.readline()
                header_names = header_line.strip("\n").split("\t")
                column_count = len(header_names)
                for column_idx in column_indices:
                    if column_idx < column_count and headers[column_idx]:
                        header_names[column_idx] = headers[column_idx]
                    else:
                        name = headers[column_idx] if headers[column_idx] else ""
                        result.message = f"Invalid column index {column_idx} with column name '{name}'"
                        return result
                with open(target_file_path, "w", encoding=write_encoding) as target:
                    self.write_row(target, header_names)
                    for line in source:
                        target.write(line)
            result.success = True
        except Exception as exc:
            result.message = f"{str(exc)}"

        return result



import logging
import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Set, Tuple, Union

from metabolights_utils.isatab.default.parser.investigation_parser import (
    parse_investigation_from_fs,
)
from metabolights_utils.isatab.default.parser.isa_table_parser import (
    assay_file_expected_patterns,
    parse_isa_table_sheet_from_fs,
    samples_file_expected_patterns,
)
from metabolights_utils.models.common import (
    CriticalMessage,
    GenericMessage,
    GenericMessageType,
)
from metabolights_utils.models.isa.assay_file import AssayFile
from metabolights_utils.models.isa.assignment_file import AssignmentFile
from metabolights_utils.models.isa.common import (
    IsaTableFile,
    OntologyItem,
    OrganismAndOrganismPartPair,
)
from metabolights_utils.models.isa.investigation_file import Investigation
from metabolights_utils.models.isa.samples_file import SamplesFile
from metabolights_utils.models.metabolights.model import (
    MetabolightsStudyModel,
    StudyDBMetadata,
    StudyFolderMetadata,
)
from metabolights_utils.models.parser.common import ParserMessage
from metabolights_utils.models.parser.enums import ParserMessageType
from metabolights_utils.provider.utils import (
    find_assay_technique,
    get_unique_file_extensions,
)

logger = logging.getLogger(__name__)


class AbstractMetadataFileProvider(ABC):

    @abstractmethod
    async def get_study_metadata_path(
        self, study_id: str, file_relative_path: Union[None, str] = None
    ) -> str:
        pass

    @abstractmethod
    async def exists(
        self, study_id: str, file_relative_path: Union[None, str] = None
    ) -> bool:
        pass


class DefaultStudyMetadataFileProvider(AbstractMetadataFileProvider):
    def __init__(self, study_metadata_root_path: str):
        self.study_metadata_root_path = study_metadata_root_path

    async def get_study_metadata_path(
        self, study_id: str, file_relative_path: Union[None, str] = None
    ) -> str:
        if file_relative_path:
            return os.path.join(
                self.study_metadata_root_path, study_id, file_relative_path
            )
        else:
            return os.path.join(self.study_metadata_root_path, study_id)

    async def exists(
        self, study_id: str, file_relative_path: Union[None, str] = None
    ) -> bool:
        file_path = Path(
            await self.get_study_metadata_path(study_id, file_relative_path)
        )
        return file_path.resolve().exists()


class AbstractDbMetadataCollector(ABC):

    @abstractmethod
    async def get_study_metadata_from_db(
        self, study_id: str, connection
    ) -> Tuple[Union[StudyDBMetadata, None], List[GenericMessage]]:
        """_summary_

        Args:
            study_id (str): MetaboLights study accession number.
            connection (Any): remote db connection to fetch study metadata.


        Returns:
            StudyDBMetadata: study metadata object that contains all the study db metadata fields.
            messages: List[GenericMessage]: list of messages to be populated with any errors or warnings.
        """


class AbstractFolderMetadataCollector(ABC):

    @abstractmethod
    async def get_folder_metadata(
        self,
        study_path,
        calculate_data_folder_size: bool = False,
        calculate_metadata_size: bool = False,
    ) -> Tuple[Union[StudyFolderMetadata, None], List[GenericMessage]]:
        """_summary_

        Args:
            study_path (_type_): MetaboLights study accession number.
            calculate_data_folder_size (bool, optional): calculate study FILES folder size. Defaults to False.
            calculate_metadata_size (bool, optional): calculate size of study ISA metadata files
                even if they are referenced or not. Defaults to False.
                If calculate_data_folder_size and calculate_metadata_size are False, folder size will be set to -1.

        Returns:
            StudyFolderMetadata: study folder metadata object that contains all descriptors of the study folders and files.
                .d, .raw and .m folders are included but not their content. If there is an error, None is returned.
            messages: List[GenericMessage]: list of messages to be populated with any errors or warnings.
        """


class AsyncMetabolightsStudyProvider(object):

    def __init__(
        self,
        db_metadata_collector: Union[None, AbstractDbMetadataCollector] = None,
        folder_metadata_collector: Union[None, AbstractFolderMetadataCollector] = None,
        metadata_file_provider: Union[None, AbstractMetadataFileProvider] = None,
    ) -> None:
        self.db_metadata_collector = db_metadata_collector
        self.folder_metadata_collector = folder_metadata_collector
        self.metadata_file_provider = metadata_file_provider

    def _add_parse_messages(
        self,
        model: MetabolightsStudyModel,
        fileName: str,
        messages: List[ParserMessage],
    ):
        if fileName not in model.parser_messages:
            model.parser_messages[fileName] = []
        model.parser_messages[fileName].extend(self.filter_messages(messages))

    def set_organisms(self, samples_file: SamplesFile, isa_table: SamplesFile):
        if (
            not samples_file
            or not samples_file.table
            or samples_file.table.total_row_count < 1
            or samples_file.table.total_column_count < 1
            or not samples_file.table.data
            or not samples_file.table.columns
        ):
            logger.warning(
                "Invalid sample file or there is no row or column in sample file. "
                + "Skip to set organisms."
            )
            return
        pairs = set()
        columns = isa_table.table.columns
        data = isa_table.table.data
        characteristics_columns = [
            "Characteristics[Organism]",
            "Characteristics[Organism part]",
            "Characteristics[Variant]",
            "Characteristics[Sample type]",
        ]
        characteristics_set = {x: set() for x in characteristics_columns}
        source_ref_column_prefix = "Term Source REF"
        accession_column_prefix = "Term Accession Number"
        characteristics_column_indices = {
            x: columns.index(x) for x in characteristics_columns if x in columns
        }

        for i in range(len(data[columns[0]])):
            characteristics: Tuple[
                OntologyItem, OntologyItem, OntologyItem, OntologyItem
            ] = (OntologyItem(), OntologyItem(), OntologyItem(), OntologyItem())
            for idx, column_name in enumerate(characteristics_columns):
                if column_name not in characteristics_column_indices:
                    continue
                if column_name in columns:
                    characteristics[idx].term = data[column_name][i]
                index = characteristics_column_indices[column_name]
                source_ref = columns[index + 1] if index + 1 < len(columns) else ""

                if source_ref.startswith(source_ref_column_prefix):
                    characteristics[idx].term_source_ref = data[columns[index + 1]][i]
                accession = columns[index + 2] if index + 2 < len(columns) else ""

                if accession.startswith(accession_column_prefix):
                    characteristics[idx].term_accession_number = data[
                        columns[index + 2]
                    ][i]
                if characteristics[idx].term and characteristics[idx].term.strip():
                    characteristics_set[column_name].add(characteristics[idx])
            if characteristics[0].term and characteristics[0].term.strip():
                pairs.add(
                    OrganismAndOrganismPartPair(
                        organism=characteristics[0],
                        organismPart=characteristics[1],
                        variant=characteristics[2],
                        sampleType=characteristics[3],
                    )
                )
        samples_file.organisms = list(characteristics_set[characteristics_columns[0]])
        samples_file.organism_parts = list(
            characteristics_set[characteristics_columns[1]]
        )
        samples_file.variants = list(characteristics_set[characteristics_columns[2]])
        samples_file.sample_types = list(
            characteristics_set[characteristics_columns[3]]
        )
        samples_file.organism_and_organism_part_pairs = list(pairs)

    async def get_file_path(
        self,
        relative_file_path: str,
        folder: Union[None, str],
        study_id: Union[None, str],
    ):
        if not folder:
            if not self.metadata_file_provider:
                raise ValueError("Define metadata file povider if folder is None.")
            file_path = await self.metadata_file_provider.get_study_metadata_path(
                study_id, relative_file_path
            )
        else:
            file_path = os.path.join(folder, relative_file_path)
        return file_path

    async def get_study_metadata_path(
        self,
        study_id: Union[None, str],
        folder: Union[None, str],
    ) -> Tuple[str, bool]:
        if not folder:
            study_path = await self.metadata_file_provider.get_study_metadata_path(
                study_id
            )
            exist = await self.metadata_file_provider.exists(study_id)
        else:
            study_path = folder
            real_path = os.path.realpath(folder)
            exist = os.path.exists(real_path)
        return study_path, exist

    async def get_phase1_input_data(
        self,
        study_id: str,
        folder: Union[str, None] = None,
        connection=None,
    ) -> MetabolightsStudyModel:
        model: MetabolightsStudyModel = MetabolightsStudyModel()
        logger.debug("Load i_Investigation.txt file on %s for %s", folder, study_id)
        await self.update_investigation_file(model, folder, study_id=study_id)
        if self.db_metadata_collector and connection:
            logger.debug("Load %s study database metadata.", study_id)
            await self.update_study_db_metadata(
                model,
                self.db_metadata_collector,
                study_id=study_id,
                connection=connection,
            )
        else:
            logger.debug("Skipped %s study database metadata.", study_id)
            message = GenericMessage(
                type=GenericMessageType.WARNING,
                short=f"Database metadata collection is skipped for {study_id}",
                detail=f"Database metadata collection is skipped for {study_id}",
            )
            model.db_reader_messages.append(message)

        if not model.investigation.studies:
            logger.warning(
                "There is no study definition in %s i_Investigation file.", study_id
            )
            return model

        raw_files = set()
        derived_files = set()
        assignment_files = set()
        folders_in_hierarchy = set()
        investigation = model.investigation
        for study_item in investigation.studies:

            file_path = await self.get_file_path(study_item.file_name, folder, study_id)

            logger.debug("Load sample file headers %s", study_item.file_name)
            samples_isa_table, messages = parse_isa_table_sheet_from_fs(
                file_path,
                samples_file_expected_patterns,
                limit=0,
                fix_unicode_exceptions=True,
                remove_empty_rows=True,
                remove_new_lines_in_cells=True,
            )
            self._add_parse_messages(model, study_item.file_name, messages)
            samples_file = SamplesFile()
            samples_file.file_path = study_item.file_name
            samples_file.table = samples_isa_table.table
            samples_file.sha256_hash = samples_isa_table.sha256_hash
            selected_column_names = set()

            selected_column_names = samples_isa_table.table.columns
            logger.debug("Load sample file rows %s", study_item.file_name)

            sample_names, message_list = parse_isa_table_sheet_from_fs(
                file_path,
                samples_file_expected_patterns,
                selected_columns=list(selected_column_names),
                fix_unicode_exceptions=True,
                remove_empty_rows=True,
                remove_new_lines_in_cells=True,
            )
            self._add_parse_messages(model, samples_file.file_path, message_list)

            samples_file.table.total_row_count = 0
            logger.debug("Find unique sample names in sample file for %s.", study_id)

            unique_sample_names = set()
            if sample_names.table.data and "Sample Name" in sample_names.table.data:
                samples_row = len(sample_names.table.data["Sample Name"])
                samples_file.table.total_row_count = samples_row

                unique_sample_names = {
                    f.strip() for f in sample_names.table.data["Sample Name"]
                }
            unique_sample_names.discard("")
            unique_sample_names.discard(None)
            samples_file.sample_names = list(unique_sample_names)
            self.set_organisms(samples_file, sample_names)

            model.samples[study_item.file_name] = samples_file
            for assay_item in study_item.study_assays.assays:
                file_path = await self.get_file_path(
                    assay_item.file_name, folder, study_id
                )

                logger.debug(
                    "Load %s assay file headers for %s.", assay_item.file_name, study_id
                )

                assay_isa_table_sheet, messages = parse_isa_table_sheet_from_fs(
                    file_path,
                    assay_file_expected_patterns,
                    limit=0,
                    fix_unicode_exceptions=True,
                    remove_empty_rows=True,
                    remove_new_lines_in_cells=True,
                )
                assay_isa_table: IsaTableFile = assay_isa_table_sheet
                assay_file: AssayFile = AssayFile()
                assay_file.file_path = assay_item.file_name
                assay_file.table = assay_isa_table.table
                assay_file.sha256_hash = assay_isa_table.sha256_hash
                self._add_parse_messages(model, assay_item.file_name, messages)
                model.assays[assay_item.file_name] = assay_file

                selected_column_names = set()
                for column in assay_isa_table.table.columns:
                    if (
                        column.startswith("Raw Spectral Data File")
                        or column.startswith("Acquisition Parameter Data File")
                        or column.startswith("Free Induction Decay Data File")
                    ):
                        selected_column_names.add(column)
                    elif column.startswith("Derived Spectral Data File"):
                        selected_column_names.add(column)
                    elif column.startswith("Metabolite Assignment File"):
                        selected_column_names.add(column)
                    elif column == "Sample Name":
                        selected_column_names.add(column)
                    elif column == "MS Assay Name" or column == "NMR Assay Name":
                        selected_column_names.add(column)
                    elif column == "Parameter Value[Column type]":
                        selected_column_names.add(column)
                logger.debug(
                    "Load %s selected assay file columns for %s.",
                    assay_item.file_name,
                    study_id,
                )
                selected_column_names_list = list(selected_column_names)
                assay_file_subset_sheet, messages_list = parse_isa_table_sheet_from_fs(
                    file_path,
                    assay_file_expected_patterns,
                    selected_columns=selected_column_names_list,
                    fix_unicode_exceptions=True,
                    remove_empty_rows=True,
                    remove_new_lines_in_cells=True,
                )
                assay_file_subset: IsaTableFile = assay_file_subset_sheet
                self._add_parse_messages(model, assay_item.file_name, messages_list)
                if selected_column_names:
                    assay_file.table.total_row_count = 0
                    if assay_file_subset.table.data:
                        assay_file.table.total_row_count = len(
                            assay_file_subset.table.data[selected_column_names_list[0]]
                        )
                else:
                    column_names = assay_file.table.columns
                    if len(column_names) > 0:
                        (
                            first_column_of_table,
                            messages_list,
                        ) = parse_isa_table_sheet_from_fs(
                            file_path,
                            assay_file_expected_patterns,
                            selected_columns=[column_names[0]],
                            fix_unicode_exceptions=True,
                            remove_empty_rows=True,
                            remove_new_lines_in_cells=True,
                        )
                        first_column: IsaTableFile = first_column_of_table
                        self._add_parse_messages(
                            model, assay_item.file_name, messages_list
                        )
                        assay_file.table.total_row_count = 0
                        if first_column.table.data:
                            assay_file.table.total_row_count = len(
                                first_column.table.data[first_column.table.columns[0]]
                            )
                    else:
                        assay_file.table.total_row_count = 0
                    assay_file.table.row_count = 0
                assay_raw_files: Set = set()
                assay_raw_file_extensions: Set = set()
                assay_derived_files: Set = set()
                assay_derived_file_extensions = set()

                for column in assay_file_subset.table.columns:
                    file_set = {f.strip() for f in assay_file_subset.table.data[column]}
                    file_set.discard("")
                    file_set.discard(None)
                    file_list = list(file_set)
                    if (
                        column.startswith("Raw Spectral Data File")
                        or column.startswith("Acquisition Parameter Data File")
                        or column.startswith("Free Induction Decay Data File")
                    ):
                        raw_files.update(file_set)
                        assay_raw_files.update(file_list)
                        assay_raw_file_extensions.update(
                            get_unique_file_extensions(file_list)
                        )
                    elif column.startswith("Derived Spectral Data File"):
                        derived_files.update(file_set)
                        assay_derived_files.update(file_list)
                        assay_derived_file_extensions.update(
                            get_unique_file_extensions(file_list)
                        )
                    elif column.startswith("Metabolite Assignment File"):
                        assignment_files.update(file_set)
                        assay_file.referenced_assignment_files = file_list
                    elif column == "Sample Name":
                        assay_file.sample_names = file_list
                    elif column == "MS Assay Name" or column == "NMR Assay Name":
                        assay_file.assay_names = file_list
                assay_file.assay_technique = find_assay_technique(
                    model, assay_file, assay_file_subset
                )

                assay_file.referenced_raw_files = list(assay_raw_files)
                assay_file.referenced_derived_files = list(assay_derived_files)
                assay_file.referenced_raw_file_extensions = list(
                    assay_raw_file_extensions
                )
                assay_file.referenced_derived_file_extensions = list(
                    assay_derived_file_extensions
                )

        for set_item in (raw_files, derived_files, assignment_files):
            for item in set_item:
                if set_item != assignment_files:
                    folders_in_hierarchy.add(os.path.dirname(item))

        model.referenced_assignment_files.extend(list(assignment_files))
        model.referenced_raw_files.extend(list(raw_files))
        model.referenced_derived_files.extend(list(derived_files))
        model.folders_in_hierarchy.extend(list(folders_in_hierarchy))

        for assignment_file in assignment_files:
            absolute_path = await self.get_file_path(assignment_file, folder, study_id)
            logger.debug(
                "Load %s assignment file  headers for %s.", assignment_file, study_id
            )
            isa_table_sheet, messages = parse_isa_table_sheet_from_fs(
                absolute_path,
                limit=0,
                fix_unicode_exceptions=True,
                remove_empty_rows=True,
                remove_new_lines_in_cells=True,
            )
            metabolite_assignment_isa_table: IsaTableFile = isa_table_sheet
            self._add_parse_messages(model, assignment_file, messages)
            file: AssignmentFile = AssignmentFile()
            file.file_path = assignment_file
            file.table = metabolite_assignment_isa_table.table
            file.sha256_hash = metabolite_assignment_isa_table.sha256_hash
            model.metabolite_assignments[assignment_file] = file
            column_names = file.table.columns

            selected_column_names = set()
            for column in file.table.columns:
                if (
                    column == "database_identifier"
                    or column == "metabolite_identification"
                ):
                    selected_column_names.add(column)

            if len(selected_column_names) > 0:
                selected_isa_table_data, messages_list = parse_isa_table_sheet_from_fs(
                    absolute_path,
                    selected_columns=list(selected_column_names),
                    fix_unicode_exceptions=True,
                    remove_empty_rows=True,
                    remove_new_lines_in_cells=True,
                )
                selected_column_values: IsaTableFile = selected_isa_table_data
                self._add_parse_messages(model, assignment_file, messages_list)

                file.table.total_row_count = 0
                metabolite_identifications = set()
                metabolite_assignments = {}
                if selected_column_values.table.data:
                    file.table.total_row_count = len(
                        selected_column_values.table.data[
                            selected_column_values.table.columns[0]
                        ]
                    )
                    for i in range(file.table.total_row_count):
                        database_identifier = ""
                        metabolite_identification = ""
                        for column in selected_column_names:
                            if column == "database_identifier":
                                database_identifier = selected_column_values.table.data[
                                    column
                                ][i]
                            elif column == "metabolite_identification":
                                metabolite_identification = (
                                    selected_column_values.table.data[column][i]
                                )
                        if metabolite_identification:
                            metabolite_identifications.add(metabolite_identification)
                            if database_identifier:
                                metabolite_assignments[metabolite_identification] = (
                                    database_identifier
                                )
                    file.identified_metabolite_names = list(metabolite_identifications)
                    file.metabolite_assignments = metabolite_assignments
                    file.number_of_assigned_rows = len(file.metabolite_assignments)
                    file.number_of_unassigned_rows = (
                        file.table.total_row_count - file.number_of_assigned_rows
                    )
                    file.number_of_rows = file.table.total_row_count
            else:
                file.table.total_row_count = 0
            file.table.row_count = 0

        for assay_file_name in model.assays:
            assay_file_item = model.assays[assay_file_name]
            for assignement_file_name in assay_file_item.referenced_assignment_files:
                model.metabolite_assignments[assignement_file_name].assay_technique = (
                    assay_file_item.assay_technique
                )
        return model

    async def get_sample_file_input(
        self,
        study_id: str,
        folder: str,
        connection,
        samples_sheet_offset: Union[int, None] = None,
        samples_sheet_limit: Union[int, None] = None,
        model: MetabolightsStudyModel = None,
    ) -> MetabolightsStudyModel:
        if not model:
            model = await self.get_phase1_input_data(study_id, folder, connection)

        for study_item in model.investigation.studies:
            file_path = await self.get_file_path(study_item.file_name, folder, study_id)
            samples_isa_table_sheet, messages = parse_isa_table_sheet_from_fs(
                file_path,
                samples_file_expected_patterns,
                offset=samples_sheet_offset,
                limit=samples_sheet_limit,
                fix_unicode_exceptions=True,
                remove_empty_rows=True,
                remove_new_lines_in_cells=True,
            )
            samples_isa_table: IsaTableFile = samples_isa_table_sheet
            model.samples[study_item.file_name].table = samples_isa_table.table
            model.samples[study_item.file_name].sha256_hash = (
                samples_isa_table.sha256_hash
            )
            table = samples_isa_table.table
            table.row_count = 0
            if table.data:
                table.row_count = len(table.data[samples_isa_table.table.columns[0]])

            model.parser_messages[study_item.file_name].extend(
                self.filter_messages(messages)
            )
            model.has_sample_table_data = True
        return model

    async def get_phase2_input_data(
        self,
        study_id: str,
        folder: Union[None, str],
        connection=None,
        samples_sheet_offset: Union[int, None] = None,
        samples_sheet_limit: Union[int, None] = None,
        assay_sheet_offset: Union[int, None] = None,
        assay_sheet_limit: Union[int, None] = None,
        model: MetabolightsStudyModel = None,
    ) -> MetabolightsStudyModel:
        if not model:
            model = await self.get_phase1_input_data(study_id, folder, connection)

        for study_item in model.investigation.studies:
            file_path = await self.get_file_path(study_item.file_name, folder, study_id)

            samples_isa_table_sheet, messages = parse_isa_table_sheet_from_fs(
                file_path,
                samples_file_expected_patterns,
                offset=samples_sheet_offset,
                limit=samples_sheet_limit,
                fix_unicode_exceptions=True,
                remove_empty_rows=True,
                remove_new_lines_in_cells=True,
            )
            samples_isa_table: IsaTableFile = samples_isa_table_sheet
            model.samples[study_item.file_name].table = samples_isa_table.table
            model.samples[study_item.file_name].sha256_hash = (
                samples_isa_table.sha256_hash
            )

            if samples_isa_table.table.data:
                model.samples[study_item.file_name].table.row_count = len(
                    samples_isa_table.table.data[samples_isa_table.table.columns[0]]
                )
            else:
                model.samples[study_item.file_name].table.row_count = 0

            model.parser_messages[study_item.file_name].extend(
                self.filter_messages(messages)
            )
            for assay_item in study_item.study_assays.assays:
                file_path = await self.get_file_path(
                    assay_item.file_name, folder, study_id
                )

                assay_isa_table_sheet, messages = parse_isa_table_sheet_from_fs(
                    file_path,
                    assay_file_expected_patterns,
                    offset=assay_sheet_offset,
                    limit=assay_sheet_limit,
                    fix_unicode_exceptions=True,
                    remove_empty_rows=True,
                    remove_new_lines_in_cells=True,
                )
                assay_isa_table: IsaTableFile = assay_isa_table_sheet
                model.parser_messages[assay_item.file_name].extend(
                    self.filter_messages(messages)
                )
                model.assays[assay_item.file_name].table = assay_isa_table.table
                model.assays[assay_item.file_name].sha256_hash = (
                    assay_isa_table.sha256_hash
                )
                model.assays[assay_item.file_name].file_path = assay_isa_table.file_path
                if assay_isa_table.table.data:
                    model.assays[assay_item.file_name].table.row_count = len(
                        assay_isa_table.table.data[assay_isa_table.table.columns[0]]
                    )
                    model.assays[assay_item.file_name].number_of_assay_rows = (
                        model.assays[assay_item.file_name].table.row_count
                    )
                else:
                    model.assays[assay_item.file_name].table.row_count = 0
        if not samples_sheet_limit or samples_sheet_limit > 0:
            model.has_sample_table_data = True
        if not assay_sheet_limit or assay_sheet_limit > 0:
            model.has_assay_table_data = True
        return model

    async def get_phase3_input_data(
        self,
        study_id: str,
        folder: str,
        connection,
        assignment_sheet_offset: Union[int, None] = None,
        assignment_sheet_limit: Union[int, None] = None,
        model: MetabolightsStudyModel = None,
    ) -> MetabolightsStudyModel:
        if not model:
            model = await self.get_phase1_input_data(study_id, folder, connection)

        for assignment_file in model.metabolite_assignments:
            absolute_path = await self.get_file_path(assignment_file, folder, study_id)
            (
                maf_isa_table_sheet,
                messages,
            ) = parse_isa_table_sheet_from_fs(
                absolute_path,
                offset=assignment_sheet_offset,
                limit=assignment_sheet_limit,
                fix_unicode_exceptions=True,
                remove_empty_rows=True,
                remove_new_lines_in_cells=True,
            )
            metabolite_assignment_isa_table: IsaTableFile = maf_isa_table_sheet
            model.parser_messages[assignment_file].extend(
                self.filter_messages(messages)
            )
            metabolite_assignment_isa_table.file_path = assignment_file
            model.metabolite_assignments[assignment_file].table = (
                metabolite_assignment_isa_table.table
            )
            model.metabolite_assignments[assignment_file].sha256_hash = (
                metabolite_assignment_isa_table.sha256_hash
            )
            model.metabolite_assignments[assignment_file].file_path = assignment_file

            if metabolite_assignment_isa_table.table.data:
                model.metabolite_assignments[assignment_file].table.row_count = len(
                    metabolite_assignment_isa_table.table.data[
                        metabolite_assignment_isa_table.table.columns[0]
                    ]
                )
            else:
                model.metabolite_assignments[assignment_file].table.row_count = 0
        if not assignment_sheet_limit or assignment_sheet_limit > 0:
            model.has_assignment_table_data = True

        return model

    async def get_phase4_input_data(
        self,
        study_id: str,
        folder: Union[str, None] = None,
        connection=None,
        model: MetabolightsStudyModel = None,
        calculate_data_folder_size: bool = False,
        calculate_metadata_size: bool = False,
    ) -> MetabolightsStudyModel:
        if not model:
            model = await self.get_phase2_input_data(study_id, folder, connection)
        try:
            metadata, messages = (
                await self.folder_metadata_collector.get_folder_metadata(
                    folder,
                    calculate_data_folder_size=calculate_data_folder_size,
                    calculate_metadata_size=calculate_metadata_size,
                )
            )

            if messages:
                model.folder_reader_messages.extend(messages)
            if metadata:
                model.study_folder_metadata = metadata
                model.has_folder_metadata = True
        except Exception as ex:
            msg = f"Error while reading folder metadata: {str(ex)}"
            model.folder_reader_messages.append(
                GenericMessage(
                    type=GenericMessageType.ERROR,
                    short=msg,
                    detail=msg,
                )
            )
        return model

    async def get_metabolights_study_model(
        self,
        study_id: str,
        folder: str,
        connection,
        load_sample_file: bool = True,
        load_assay_files: bool = True,
        load_maf_files: bool = True,
        load_folder_metadata: bool = True,
        samples_sheet_offset: Union[int, None] = None,
        samples_sheet_limit: Union[int, None] = None,
        assay_sheet_offset: Union[int, None] = None,
        assay_sheet_limit: Union[int, None] = None,
        assignment_sheet_offset: Union[int, None] = None,
        assignment_sheet_limit: Union[int, None] = None,
        calculate_data_folder_size: bool = False,
        calculate_metadata_size: bool = False,
    ) -> MetabolightsStudyModel:
        return await self.load_study(
            study_id,
            folder,
            connection,
            load_sample_file=load_sample_file,
            load_assay_files=load_assay_files,
            load_maf_files=load_maf_files,
            load_folder_metadata=load_folder_metadata,
            samples_sheet_offset=samples_sheet_offset,
            samples_sheet_limit=samples_sheet_limit,
            assay_sheet_offset=assay_sheet_offset,
            assay_sheet_limit=assay_sheet_limit,
            assignment_sheet_offset=assignment_sheet_offset,
            assignment_sheet_limit=assignment_sheet_limit,
            calculate_data_folder_size=calculate_data_folder_size,
            calculate_metadata_size=calculate_metadata_size,
        )

    async def load_study(
        self,
        study_id: str,
        study_path: str,
        connection=None,
        load_sample_file: bool = False,
        load_assay_files: bool = False,
        load_maf_files: bool = False,
        load_folder_metadata: bool = False,
        samples_sheet_offset: Union[int, None] = None,
        samples_sheet_limit: Union[int, None] = None,
        assay_sheet_offset: Union[int, None] = None,
        assay_sheet_limit: Union[int, None] = None,
        assignment_sheet_offset: Union[int, None] = None,
        assignment_sheet_limit: Union[int, None] = None,
        calculate_data_folder_size: bool = False,
        calculate_metadata_size: bool = False,
    ) -> MetabolightsStudyModel:
        if not study_id:
            raise ValueError("invalid study_id")
        exist = False
        study_path, exist = await self.get_study_metadata_path(study_id, study_path)

        if not study_path:
            raise ValueError("invalid study_path")
        if not exist:
            model = MetabolightsStudyModel()
            message = CriticalMessage(
                short=f"Study folder does not exist for {study_id} {study_path}"
            )
            model.folder_reader_messages.append(message)
            return model

        model = await self.get_phase1_input_data(study_id, study_path, connection)
        if load_sample_file and not load_assay_files and not load_folder_metadata:
            model = await self.get_sample_file_input(
                study_id,
                study_path,
                connection,
                model=model,
                samples_sheet_offset=samples_sheet_offset,
                samples_sheet_limit=samples_sheet_limit,
            )

        if load_assay_files or load_folder_metadata:
            model = await self.get_phase2_input_data(
                study_id,
                study_path,
                connection,
                model=model,
                samples_sheet_offset=samples_sheet_offset,
                samples_sheet_limit=samples_sheet_limit,
                assay_sheet_offset=assay_sheet_offset,
                assay_sheet_limit=assay_sheet_limit,
            )

        if load_folder_metadata:
            model = await self.get_phase4_input_data(
                study_id,
                study_path,
                connection=connection,
                model=model,
                calculate_data_folder_size=calculate_data_folder_size,
                calculate_metadata_size=calculate_metadata_size,
            )
        else:
            message = GenericMessage(
                type=GenericMessageType.WARNING,
                short=f"Folder metadata collection is skipped for {study_id}",
                detail=f"Folder metadata collection is skipped for {study_id}",
            )
            model.folder_reader_messages.append(message)

        if load_maf_files:
            model = await self.get_phase3_input_data(
                study_id,
                study_path,
                connection,
                model=model,
                assignment_sheet_limit=assignment_sheet_limit,
                assignment_sheet_offset=assignment_sheet_offset,
            )

        return model

    async def update_investigation_file(
        self,
        model: MetabolightsStudyModel,
        folder,
        file_name="i_Investigation.txt",
        study_id: Union[None, str] = None,
    ):
        file = await self.get_file_path(file_name, folder, study_id)
        investigation, messages = parse_investigation_from_fs(
            file, fix_unicode_exceptions=True
        )
        if not investigation:
            investigation = Investigation()

        model.investigation_file_path = file_name
        model.investigation = investigation
        model.parser_messages[file_name] = self.filter_messages(messages)
        model.has_investigation_data = True
        return model

    async def update_study_db_metadata(
        self,
        model: MetabolightsStudyModel,
        collector: AbstractDbMetadataCollector,
        study_id,
        connection,
    ):
        try:
            study_db_metadata, messages = await collector.get_study_metadata_from_db(
                study_id, connection
            )
            if messages:
                model.db_reader_messages.extend(messages)
            if study_db_metadata:
                model.study_db_metadata = study_db_metadata
                model.has_db_metadata = True
        except Exception as ex:
            logger.exception("Error while reading datadata for %s", study_id)
            message = GenericMessage(
                type=GenericMessageType.ERROR,
                short=f"Error while reading database for {study_id} metadata",
                detail=f"{str(ex)}",
            )
            model.db_reader_messages.append(message)
        return model

    def filter_messages(self, messages: List[ParserMessage]):
        return [
            f
            for f in messages
            if f.type == ParserMessageType.CRITICAL
            or f.type == ParserMessageType.ERROR
            or f.type == ParserMessageType.WARNING
        ]



import datetime
import glob
import io
import json
import os
import time
from typing import List, Literal, Tuple, Union

import httpx

from metabolights_utils.commands.submission.model import (
    FtpLoginCredentials,
    RestApiCredentials,
    StudyResponse,
    SubmittedStudiesResponse,
)
from metabolights_utils.commands.submission.utils import (
    get_submission_private_ftp_credentials,
    get_submission_rest_api_credentials,
)
from metabolights_utils.models.common import ErrorMessage, GenericMessage, InfoMessage
from metabolights_utils.models.enums import GenericMessageType
from metabolights_utils.models.metabolights.model import (
    MetabolightsStudyModel,
    StudyFolderMetadata,
)
from metabolights_utils.provider import definitions
from metabolights_utils.provider.ftp.default_ftp_client import (
    DefaultFtpClient,
    LocalDirectory,
)
from metabolights_utils.provider.ftp.folder_metadata_collector import (
    FtpFolderMetadataCollector,
)
from metabolights_utils.provider.local_folder_metadata_collector import (
    LocalFolderMetadataCollector,
)
from metabolights_utils.provider.study_provider import (
    AbstractDbMetadataCollector,
    AbstractFolderMetadataCollector,
    MetabolightsStudyProvider,
)
from metabolights_utils.provider.submission_model import (
    APIResponse,
    FtpUploadDetails,
    PolicyResultResponse,
    ValidationMessage,
    ValidationReport,
    ValidationResponse,
    WorkerTaskStatus,
)
from metabolights_utils.provider.utils import (
    download_file_from_rest_api,
    is_metadata_file,
    is_metadata_filename_pattern,
    rest_api_get,
)
from metabolights_utils.utils.filename_utils import join_path


class MetabolightsSubmissionRepository:
    def __init__(
        self,
        local_storage_root_path: Union[None, str] = None,
        credentials_file_path: Union[None, str] = None,
        ftp_server_url: Union[None, str] = None,
        rest_api_base_url: Union[None, str] = None,
        validation_api_base_url: Union[None, str] = None,
        local_storage_cache_path: Union[None, str] = None,
    ) -> None:
        self.ftp_server_url = ftp_server_url
        if not self.ftp_server_url:
            self.ftp_server_url = definitions.default_private_ftp_server_url

        self.local_storage_cache_path = local_storage_cache_path
        if not local_storage_cache_path:
            self.local_storage_cache_path = (
                definitions.default_local_submission_cache_path
            )
        self.local_storage_root_path = local_storage_root_path
        if not self.local_storage_root_path:
            self.local_storage_root_path = (
                definitions.default_local_submission_root_path
            )

        self.credentials_file_path = credentials_file_path
        if not self.credentials_file_path:
            self.credentials_file_path = (
                definitions.default_local_submission_credentials_file_path
            )
        self.rest_api_base_url = rest_api_base_url
        if not self.rest_api_base_url:
            self.rest_api_base_url = definitions.default_rest_api_url
        self.validation_api_base_url = validation_api_base_url
        if not self.validation_api_base_url:
            self.validation_api_base_url = definitions.default_validation_api_url

    def load_study_model(
        self,
        study_id: str,
        local_path: Union[str, None] = None,
        use_only_local_path: bool = False,
        override_local_files: bool = False,
        load_folder_metadata: bool = True,
        rebuild_folder_index_file: bool = False,
        folder_index_file_path: Union[str, None] = None,
        db_metadata_collector: Union[None, AbstractDbMetadataCollector] = None,
        folder_metadata_collector: Union[None, AbstractFolderMetadataCollector] = None,
    ) -> Tuple[Union[None, MetabolightsStudyModel], List[GenericMessage]]:
        if not study_id or not study_id.strip():
            return None, [ErrorMessage(short="Invalid study_id")]
        study_id = study_id.upper().strip("/")
        if not local_path:
            local_path = self.local_storage_root_path
        target_path = os.path.join(local_path, study_id)
        if not folder_index_file_path:
            folder_index_file_path = os.path.join(
                self.local_storage_cache_path,
                study_id,
                "mtbls_index.json",
            )

        model: Union[None, MetabolightsStudyModel] = None
        messages = []
        if use_only_local_path:
            provider = MetabolightsStudyProvider(
                db_metadata_collector=db_metadata_collector,
                folder_metadata_collector=folder_metadata_collector,
            )
            model: MetabolightsStudyModel = provider.load_study(
                study_id,
                study_path=target_path,
                connection=None,
                load_assay_files=True,
                load_sample_file=True,
                load_maf_files=True,
                load_folder_metadata=load_folder_metadata,
                calculate_data_folder_size=True,
                calculate_metadata_size=True,
            )

            return model, [InfoMessage(short="Loaded from local isa metadata files.")]

        messages: List[GenericMessage] = []
        try:
            result = self.download_submission_metadata_files(
                study_id=study_id,
                local_path=local_path,
                metadata_files=None,
                override_local_files=override_local_files,
                delete_unlisted_metadata_files=True,
            )
            if result.success:

                messages.append(
                    InfoMessage(
                        short="Downloaded metadata file with response",
                        detail=f"Response message: {result.code} {result.message}",
                    )
                )

                provider = MetabolightsStudyProvider(
                    db_metadata_collector=db_metadata_collector,
                    folder_metadata_collector=LocalFolderMetadataCollector(),
                )
                model: MetabolightsStudyModel = provider.load_study(
                    study_id,
                    study_path=target_path,
                    connection=None,
                    load_assay_files=True,
                    load_sample_file=True,
                    load_maf_files=True,
                    load_folder_metadata=load_folder_metadata,
                    calculate_data_folder_size=False,
                    calculate_metadata_size=False,
                )
            else:
                messages.append(
                    ErrorMessage(
                        short="Download metadata file failure.",
                        detail=f"Error message: {result.code} {result.message}",
                    )
                )

        except Exception as ex:
            messages.append(
                ErrorMessage(
                    short="Download metadata file failure. Try -l option if there is a local copy.",
                    detail=str(ex),
                )
            )
        return model, messages

    def upload_metadata_files(
        self,
        study_id,
        remote_folder_directory: Union[str, None] = None,
        local_path: Union[str, None] = None,
        ftp_server_url: Union[str, None] = None,
        metadata_files: Union[List[str], None] = None,
        override_remote_files: bool = False,
        ftp_username: Union[str, None] = None,
        ftp_password: Union[str, None] = None,
        user_api_token: Union[str, None] = None,
    ) -> Tuple[bool, str]:
        if not local_path:
            local_path = self.local_storage_root_path
        local_path = join_path(local_path)

        if not ftp_server_url:
            ftp_server_url = self.ftp_server_url

        response, errors = self.list_isa_metadata_files(
            study_id=study_id, user_api_token=user_api_token
        )
        if not response:
            return False, "Errors while listing metadata files."
        if not remote_folder_directory:
            if response.upload_path:
                remote_folder_directory = response.upload_path
            else:
                return False, "Remote folder does not defined."
        modified_time_dict = {}
        new_requested_files = []
        for descriptor in response.study:
            _time = descriptor.created_at
            try:
                modified = datetime.datetime.strptime(
                    _time, "%Y-%m-%d %H:%M:%S"
                ).timestamp()
            except ValueError:
                modified = 0
            remote_modified_time = int(modified)
            modified_time_dict[descriptor.file] = remote_modified_time
        study_path = os.path.join(local_path, study_id)
        study_path = os.path.realpath(study_path)
        if not os.path.exists(study_path):
            return False, f"Study path does not exist: {study_path}"
        if not metadata_files or override_remote_files:
            files = os.listdir(study_path)
            metadata_files = [x for x in files if is_metadata_filename_pattern(x)]
        if override_remote_files:
            new_requested_files = metadata_files
        else:
            for file_name in metadata_files:
                relative_path = file_name.replace(f"{local_path}/", "", 1)
                remote_modified_time = None
                if relative_path in modified_time_dict:
                    remote_modified_time = modified_time_dict[relative_path]
                file_path = os.path.join(study_path, file_name)
                if remote_modified_time:
                    local_modified_time = int(os.path.getmtime(file_path))
                    if remote_modified_time < local_modified_time:
                        new_requested_files.append(relative_path)
                else:
                    new_requested_files.append(relative_path)
        if not new_requested_files:
            return (
                False,
                "There is no metadata file to upload or local metadata files are up-to-date.",
            )
        if not ftp_password or not ftp_username:
            ftp_username, ftp_password, error = self.get_ftp_credentials()

        ftp_client = DefaultFtpClient(
            local_storage_root_path=local_path,
            ftp_server_url=ftp_server_url,
            remote_repository_root_directory="",
            username=ftp_username,
            password=ftp_password,
        )
        input_files = [os.path.join(study_path, x) for x in new_requested_files]
        try:
            success, message = ftp_client.upload_files(
                remote_folder_directory, input_files
            )
            if success:
                return success, "Uploaded Files: " + ", ".join(new_requested_files)
            else:
                return success, message
        except Exception as ex:
            return False, str(ex)

    def upload_data_files(
        self,
        study_id,
        local_path: Union[str, None] = None,
        remote_folder_directory: Union[str, None] = None,
        ftp_server_url: Union[str, None] = None,
        ftp_username: Union[str, None] = None,
        ftp_password: Union[str, None] = None,
    ) -> Tuple[bool, str]:
        if not local_path:
            local_path = self.local_storage_root_path
        local_study_path = join_path(local_path, study_id)
        if not ftp_server_url:
            ftp_server_url = self.ftp_server_url

        study_path = os.path.realpath(local_study_path)
        if not os.path.exists(study_path):
            return False, f"Study path does not exist: {study_path}"

        files = glob.iglob(f"{study_path}/**/*", recursive=True)

        input_files = [
            x for x in files if not is_metadata_filename_pattern(os.path.basename(x))
        ]

        # username, password, error = self.get_ftp_credentials()
        ftp_client = DefaultFtpClient(
            local_storage_root_path=local_path,
            ftp_server_url=ftp_server_url,
            remote_repository_root_directory="",
            username=ftp_username,
            password=ftp_password,
        )
        try:
            ftp_client.upload_files(remote_folder_directory, input_files)
            return True, "Uploaded Files: " + ", ".join(input_files)
        except Exception as ex:
            return False, str(ex)

    def check_api_response(self, api_name, response):
        if not response:
            return False, f"No response for {api_name}"
        code = response.status_code
        text = response.text
        if response.status_code in (200, 201):
            return True, None
        else:
            return (
                False,
                f"Failure of {api_name} {code}: {text}",
            )

    def get_ftp_upload_details(
        self,
        study_id,
        rest_api_base_url: Union[str, None] = None,
        user_api_token: Union[str, None] = None,
        timeout: int = 5,
    ) -> Tuple[Union[None, FtpUploadDetails], str]:

        headers = {}
        if not user_api_token:
            user_api_token, error = self.get_api_token()
            if not user_api_token:
                return None, error
        sub_path = f"/studies/{study_id}/upload-info"
        headers["User-Token"] = user_api_token

        api_name = "get private ftp upload details"
        rest_api_base_url = (
            rest_api_base_url if rest_api_base_url else self.rest_api_base_url
        )
        try:
            url = f"{rest_api_base_url.rstrip('/')}/{sub_path.lstrip('/')}"
            parameters = {}
            response = httpx.get(
                url=url,
                timeout=timeout,
                headers=headers,
                params=parameters,
            )
            success, error = self.check_api_response(api_name, response)
            if not success:
                return None, error

            data = json.loads(response.text)
            details = FtpUploadDetails.model_validate(data, from_attributes=True)
            return details, ""
        except Exception as ex:
            return None, f"Validation task start failure: {str(ex)}."

    def validate_study(
        self,
        study_id,
        validation_file_path: Union[str, None] = None,
        pool_period: int = 5,
        retry: int = 20,
        timeout: int = 10,
    ):
        if not validation_file_path:
            validation_file_path = os.path.join(
                self.local_storage_cache_path, study_id, f"{study_id}_validation.tsv"
            )
        validation_file_path = os.path.realpath(validation_file_path)

        sub_path = f"/studies/{study_id}/validation-task"

        api_header, error = self.get_api_token()
        headers = {}
        if api_header:
            headers["User-Token"] = api_header
        else:
            return None, error

        api_name = "validation task start"
        try:
            url = f"{self.rest_api_base_url.rstrip('/')}/{sub_path.lstrip('/')}"
            parameters = {}
            response = httpx.post(
                url=url,
                timeout=timeout,
                headers=headers,
                params=parameters,
            )
            success, error = self.check_api_response(api_name, response)
            if not success:
                return None, error

            data = json.loads(response.text)
            task_start_response = ValidationResponse.model_validate(
                data, from_attributes=True
            )

            if not task_start_response.task.task_id:
                return (None, f"Failure of {api_name} for {study_id}.")

        except Exception as ex:
            return None, f"Validation task start failure: {str(ex)}."

        api_name = "validation task status check"
        try:
            task_success = False
            for _ in range(retry + 1):
                time.sleep(pool_period)
                try:
                    response = httpx.get(
                        url=url,
                        timeout=timeout,
                        headers=headers,
                        params=parameters,
                    )
                except TimeoutError:
                    continue
                success, error = self.check_api_response(api_name, response)
                if not success:
                    continue
                data = json.loads(response.text)
                task_status_response = ValidationResponse.model_validate(
                    data, from_attributes=True
                )

                task = task_status_response.task
                status = task.last_status
                if task.task_id and task.last_status:
                    if "SUCCESS" in status.upper():
                        task_success = True
                        break
            if not task_success:
                return (None, f"Failure of {api_name} for {study_id}.")
        except Exception as ex:
            return None, f"{api_name} failure: {str(ex)}."

        api_name = "get validation report"
        try:
            api_success = False
            report_sub_path = f"/studies/{study_id}/validation-report"
            report_url = (
                f"{self.rest_api_base_url.rstrip('/')}/{report_sub_path.lstrip('/')}"
            )
            for _ in range(retry + 1):
                try:
                    response = httpx.get(
                        url=report_url,
                        timeout=timeout,
                        headers=headers,
                        params=parameters,
                    )
                except TimeoutError:
                    time.sleep(pool_period)
                    continue
                success, error = self.check_api_response(api_name, response)
                if not success:
                    time.sleep(pool_period)
                    continue
                data = json.loads(response.text)
                report = ValidationReport.model_validate(data, from_attributes=True)
                if report.validation.status:
                    api_success = True
                    break
            if not api_success:
                return (None, f"Failure of {api_name} for {study_id}.")
        except Exception as ex:
            return None, f"{api_name} failure: {str(ex)}."

        errors: List[ValidationMessage] = []
        for section in report.validation.validations:
            for message in section.details:
                if message.status.upper() == "ERROR":
                    errors.append(message)
        dir_name = os.path.dirname(validation_file_path)
        os.makedirs(dir_name, exist_ok=True)

        with open(validation_file_path, "w", encoding="utf-8") as f:
            f.write(
                "section\t" "status\t" "message\t" "description\t" "metadata_file\n"
            )
            for e in errors:
                f.write(
                    f"{e.section}\t"
                    f"{e.status}\t"
                    f"{e.message}\t"
                    f"{e.description}\t"
                    f"{e.metadata_file}\n"
                )
        return True, status

    def validate_study_v2(
        self,
        study_path: str,
        validation_result_file_path: str,
        pool_period: int = 5,
        retry: int = 20,
        timeout: int = 30,
        validation_api_base_url: Union[None, str] = None,
        rest_api_base_url: Union[None, str] = None,
        api_token: Union[None, str] = None,
    ):

        sub_path = f"/study-model/validation"
        auth_sub_path = "/auth/login-with-token"
        provider = MetabolightsStudyProvider(
            db_metadata_collector=None,
            folder_metadata_collector=LocalFolderMetadataCollector(),
        )
        basename = os.path.basename(study_path)
        model: MetabolightsStudyModel = provider.load_study(
            basename,
            study_path=study_path,
            connection=None,
            load_assay_files=True,
            load_sample_file=True,
            load_maf_files=True,
            load_folder_metadata=True,
            calculate_data_folder_size=False,
            calculate_metadata_size=False,
        )
        api_header = api_token
        if not api_token:
            api_header, error = self.get_api_token()
        api_name = "validation v2 get jwt token"
        jwt_token = None
        rest_api_base_url = (
            rest_api_base_url if rest_api_base_url else self.rest_api_base_url
        )

        try:
            url = f"{rest_api_base_url.rstrip('/')}/{auth_sub_path.lstrip('/')}"
            parameters = {}
            response = httpx.post(
                url=url,
                timeout=timeout,
                json={"token": api_header},
                headers={},
                params=parameters,
            )
            success, error = self.check_api_response(api_name, response)
            if not success:
                return None, error
            if "jwt" in response.headers:
                jwt_token = response.headers["jwt"]

            if not jwt_token:
                return (None, "Failure of authentication.")

        except Exception as ex:
            return None, f"Validation task v2 authentication failure: {str(ex)}."

        study_model_str = model.model_dump_json(by_alias=True)
        json_file = io.BytesIO(study_model_str.encode("utf-8"))

        files = {"study_model_json": ("study_model.json", json_file)}

        headers = {"accept": "application/json"}
        headers["Authorization"] = f"Bearer {jwt_token}"
        task_id = None
        api_name = "validation v2 task start"
        validation_api_base_url = (
            validation_api_base_url
            if validation_api_base_url
            else self.validation_api_base_url
        )
        try:
            url = f"{validation_api_base_url.rstrip('/')}/{sub_path.lstrip('/')}"
            parameters = {}
            response = httpx.post(
                url=url,
                timeout=timeout,
                headers=headers,
                files=files,
                params=parameters,
            )
            success, error = self.check_api_response(api_name, response)
            if not success:
                return None, error

            data = json.loads(response.text)
            task_start_response = APIResponse[WorkerTaskStatus].model_validate(
                data, from_attributes=True
            )

            if not task_start_response.content.task_id:
                return (None, f"Failure of {api_name}.")
            task_id = task_start_response.content.task_id
        except Exception as ex:
            return None, f"Validation task v2 start failure: {str(ex)}."

        headers["task_id"] = task_id
        api_name = "validation task v2 status check"
        try:
            task_success = False
            for _ in range(retry + 1):
                time.sleep(pool_period)
                try:
                    response = httpx.get(
                        url=url,
                        timeout=timeout,
                        headers=headers,
                        params=parameters,
                    )
                except TimeoutError:
                    continue
                success, error = self.check_api_response(api_name, response)
                if not success:
                    continue
                data = json.loads(response.text)
                task_status_response = APIResponse[PolicyResultResponse].model_validate(
                    data, from_attributes=True
                )

                task = task_status_response.content
                status = task.task_status
                if task.task_id and status:
                    if "SUCCESS" in status.upper():
                        task_success = True
                        break
            if not task_success:
                return None, f"Failure of {api_name}."
        except Exception as ex:
            return None, f"{api_name} failure: {str(ex)}."

        result = task_status_response.content.task_result
        with open(validation_result_file_path, "w", encoding="utf-8") as f:
            json.dump(result.model_dump(by_alias=True), f, indent=4)

        return True, task_status_response.content.task_id

    def sync_private_ftp_metadata_files(
        self,
        study_id: str,
        pool_period: int = 10,
        retry: int = 10,
        timeout: int = 10,
        user_api_token: Union[None, str] = None,
    ):
        return self.sync_from_private_ftp(
            study_id,
            "metadata",
            pool_period,
            retry,
            timeout,
            user_api_token=user_api_token,
        )

    def sync_private_ftp_data_files(
        self,
        study_id: str,
        pool_period: int = 10,
        retry: int = 1000,
        timeout: int = 10,
        user_api_token: Union[None, str] = None,
    ):
        return self.sync_from_private_ftp(
            study_id, "data", pool_period, retry, timeout, user_api_token=user_api_token
        )

    def sync_from_private_ftp(
        self,
        study_id: str,
        sync_type: Literal["metadata", "data"] = "metadata",
        pool_period: int = 10,
        retry: int = 10,
        timeout: int = 10,
        user_api_token: Union[None, str] = None,
    ):
        sub_path = f"/studies/{study_id}/study-folders/rsync-task"
        if not user_api_token:
            user_api_token, error = self.get_api_token()
        headers = {}
        if user_api_token:
            headers["User-Token"] = user_api_token
        else:
            return None, error
        headers["Dry-Run"] = "false"
        headers["Sync-Type"] = sync_type
        headers["Source-Staging-Area"] = "private-ftp"
        if sync_type == "metadata":
            headers["Target-Staging-Area"] = "rw-study"
        else:
            headers["Target-Staging-Area"] = "readonly-study"

        url = f"{self.rest_api_base_url.rstrip('/')}/{sub_path.lstrip('/')}"
        parameters = {}
        response = httpx.post(
            url=url,
            timeout=timeout,
            headers=headers,
            params=parameters,
        )
        task_id = None
        if response and response.status_code in (200, 201):
            data = json.loads(response.text)
            if "task_id" in data:
                task_id = data["task_id"]
            # elif "task" in data and "task_id" in data["task"]:
            #     task_id = data["task"]["task_id"]

            if task_id:
                for _ in range(retry + 1):
                    time.sleep(pool_period)
                    response = httpx.get(
                        url=url,
                        timeout=timeout,
                        headers=headers,
                        params=parameters,
                    )
                    if response:
                        data = json.loads(response.text)
                        if "status" in data:
                            status = data["status"]
                            if "SUCCESS" in status:
                                return True, status
                            elif "FAIL" in status:
                                return False, status
                return False, "No result"
            else:
                return False, "Response does not have a task id"
        else:
            return (
                (False, response.text)
                if response
                else (False, f"Response code: {str(response.status_code)}")
            )

    def create_submission(
        self,
        user_api_token: Union[None, str] = None,
        timeout: int = 30,
    ) -> Tuple[Union[None, str], Union[None, str]]:
        if not user_api_token:
            user_api_token, error = self.get_api_token()
        headers = {}
        if user_api_token:
            headers["User-Token"] = user_api_token
        else:
            return None, error

        sub_path = f"/studies/create"
        url = f"{self.rest_api_base_url.rstrip('/')}/{sub_path.lstrip('/')}"

        try:
            response = httpx.get(url=url, timeout=timeout, headers=headers, params={})
            if response and response.status_code in (200, 201):
                data = json.loads(response.text)
                if "new_study" in data:
                    return data["new_study"], ""

                return None, "Invalid response."
            else:
                return None, response.text if response else "Invalid response."
        except Exception as ex:
            return None, str(ex)

    def create_assay(
        self,
        study_id: str,
        assay_technique: str,
        scan_polarity: str = "",
        column_type: str = "",
        timeout: int = 30,
    ):
        if not study_id or not assay_technique:
            return None, None, "study_id or assay_technique is not defined."

        api_header, error = self.get_api_token()
        headers = {}
        if api_header:
            headers["User-Token"] = api_header
        else:
            return None, None, error

        sub_path = f"/studies/{study_id}/assays"
        url = f"{self.rest_api_base_url.rstrip('/')}/{sub_path.lstrip('/')}"
        body = {"assay": {"type": assay_technique, "columns": []}}
        if scan_polarity:
            body["assay"]["columns"].append(
                {"name": "polarity", "value": scan_polarity.lower()}
            )
        if column_type:
            body["assay"]["columns"].append(
                {"name": "column type", "value": column_type.lower()}
            )
        try:
            response = httpx.post(
                url=url, timeout=timeout, headers=headers, params={}, json=body
            )
            if response and response.status_code in (200, 201):
                data = json.loads(response.text)
                if "filename" in data and "maf" in data:
                    return data["filename"], data["maf"], None

                return None, None, "Invalid response."
            else:
                return None, None, response.status_code if response else None
        except Exception as ex:
            return None, None, str(ex)

    def delete_assay(
        self,
        study_id: str,
        assay_filename: str,
        save_audit_copy: bool = True,
        timeout: int = 20,
    ):
        if not study_id or not assay_filename:
            return None, None, "study_id or assay_filename is not defined."

        api_header, error = self.get_api_token()
        headers = {"save_audit_copy": str(save_audit_copy).lower()}
        if api_header:
            headers["User-Token"] = api_header
        else:
            return None, None, error

        sub_path = f"/studies/{study_id}/assays/{assay_filename}"
        url = f"{self.rest_api_base_url.rstrip('/')}/{sub_path.lstrip('/')}"
        try:
            response = httpx.delete(
                url=url,
                timeout=timeout,
                headers=headers,
                params={},
            )
            if response and response.status_code == 404:
                return False, f"{assay_filename} does not exist."
            if response and response.status_code in (200, 201):
                data = json.loads(response.text)
                if "success" in data:
                    return True, None

                return False, "Invalid response."
            else:
                return False, response.text if response else None
        except Exception as ex:
            return False, str(ex)

    def download_submission_metadata_files(
        self,
        study_id: str,
        local_path: Union[str, None] = None,
        metadata_files: Union[List[str], None] = None,
        override_local_files: bool = False,
        delete_unlisted_metadata_files: bool = True,
    ) -> LocalDirectory:

        api_header, error = self.get_api_token()
        headers = {}
        if api_header:
            headers["User-Token"] = api_header
        else:
            return LocalDirectory(code=400, message=f"user token error: {error}")

        if not study_id:
            return LocalDirectory(code=400, message="Invalid study_id")
        if not local_path:
            local_path = self.local_storage_root_path
        local_path = join_path(local_path)
        response = LocalDirectory(root_path=local_path)

        study_id = study_id.upper().strip("/")

        listed_files = []
        requested_files = metadata_files
        if not metadata_files or delete_unlisted_metadata_files:
            result, error = self.list_isa_metadata_files(study_id)
            if result:
                file_names = [x.file for x in result.study]
                descriptors = {x.file: x for x in result.study}
                listed_files = set(file_names)
                if not metadata_files:
                    requested_files = file_names
            else:
                return LocalDirectory(
                    code=404, message="There is no metadata file to download."
                )

        if requested_files != metadata_files:
            filtered_files = [
                x for x in requested_files if is_metadata_filename_pattern(x)
            ]
            requested_files = filtered_files
        response.actions = {f"{study_id}/{x}": "SKIPPED" for x in requested_files}

        new_requested_files = []
        for filename in requested_files:
            file_path = os.path.join(local_path, study_id, filename)
            key = f"{study_id}/{filename}"
            if os.path.exists(file_path):
                if override_local_files:
                    response.actions[key] = "OVERRIDED"
                else:
                    local_modified_time = int(os.path.getmtime(file_path))
                    _time = descriptors[filename].created_at
                    try:
                        modified = datetime.datetime.strptime(
                            _time, "%Y-%m-%d %H:%M:%S"
                        ).timestamp()
                    except ValueError:
                        modified = 0
                    remote_modified_time = int(modified)
                    if remote_modified_time > local_modified_time:
                        new_requested_files.append(filename)
                        response.actions[key] = "DOWNLOADED"
                    else:
                        response.actions[key] = "SKIPPED"
            else:
                new_requested_files.append(filename)
                response.actions[key] = "DOWNLOADED"
        requested_files = new_requested_files

        if requested_files and not os.path.exists(local_path):
            os.makedirs(local_path, exist_ok=True)

        messages: List[GenericMessage] = []
        current_file = None
        try:
            sub_path = f"/studies/{study_id}/download"

            if not metadata_files:
                parameters = {"file": "metadata"}
            else:
                parameters["file"] = [",".join(metadata_files)]

            for filename in requested_files:
                new_file_path = os.path.join(local_path, study_id, filename)
                current_file = filename
                url = self.rest_api_base_url.strip("/") + "/" + sub_path.strip("/")
                try:
                    _time = descriptors[filename].created_at
                    try:
                        modified = datetime.datetime.strptime(
                            _time, "%Y-%m-%d %H:%M:%S"
                        ).timestamp()
                    except ValueError:
                        modified = 0
                    remote_modified_time = int(modified)
                except Exception:
                    remote_modified_time = None

                success = download_file_from_rest_api(
                    url,
                    new_file_path,
                    timeout=60,
                    headers=headers,
                    parameters=parameters,
                    modification_time=remote_modified_time,
                    is_zip_response=True,
                )
                if not success:
                    response.actions[filename] = "FAILED"
            if delete_unlisted_metadata_files:
                study_path = os.path.join(local_path, study_id)
                if os.path.exists(study_path):
                    for filename in os.listdir(study_path):
                        if filename not in listed_files:
                            file_path = os.path.join(study_path, filename)
                            if is_metadata_file(file_path):
                                response.actions[f"{study_id}/{filename}"] = "DELETED"
                                os.remove(file_path)
            response.success = True
            response.code = 200
            response.message = "Ok"

        except Exception as ex:
            messages.append(
                ErrorMessage(
                    short=f"Download data file {current_file if current_file else ''} failure",
                    detail=f"Error message: {str(ex)}",
                )
            )

        return response

    def list_isa_metadata_files(
        self,
        study_id: str,
        user_api_token: Union[str, None] = None,
    ) -> Tuple[Union[StudyResponse, None], Union[None, str]]:

        response, error = self.list_study_directory(
            study_id=study_id, user_api_token=user_api_token
        )
        if response:
            response.study = [
                x for x in response.study if is_metadata_filename_pattern(x.file)
            ]
            return response, None
        else:
            return response, error

    def list_studies(
        self, timeout=None, headers=None, parameters=None
    ) -> Tuple[Union[None, SubmittedStudiesResponse], Union[str, None]]:
        api_header, error = self.get_api_token()
        headers = {}
        if api_header:
            headers["User-Token"] = api_header
        else:
            return None, error

        sub_path = "/studies/user"
        url = f"{self.rest_api_base_url.rstrip('/')}/{sub_path.lstrip('/')}"
        response = httpx.get(
            url=url,
            timeout=timeout,
            headers=headers,
            params=parameters,
        )
        if response and response.status_code in (200, 201):
            data = json.loads(response.text)
            studies_data = SubmittedStudiesResponse.model_validate(data)
            studies = studies_data.data.copy()
            studies.sort(key=lambda x: x.updated)

            return studies_data, None
        else:
            return None, response.status_code if response else None

    def get_api_token(self):
        result = None
        if self.credentials_file_path:
            credentials: RestApiCredentials = get_submission_rest_api_credentials(
                credentials_file_path=self.credentials_file_path,
                rest_api_base_url=self.rest_api_base_url,
            )
            result = credentials.api_token if credentials else None

        if not result:
            return (
                None,
                f"There is not user api token for {self.rest_api_base_url} "
                "Rest API. Login before to use the command.",
            )
        return (result, None)

    def get_ftp_credentials(self):
        result = None
        if self.credentials_file_path:
            credentials: FtpLoginCredentials = get_submission_private_ftp_credentials(
                credentials_file_path=self.credentials_file_path,
                private_ftp_url=self.ftp_server_url,
            )
            result = credentials if credentials else None

        if not result:
            return (
                None,
                None,
                f"There is not user api token for {self.rest_api_base_url} "
                "Rest API. Login before to use the command.",
            )
        return (result.user_name, result.password, None)

    def list_study_directory(
        self,
        study_id: str,
        subdirectory: Union[str, None] = None,
        timeout=None,
        user_api_token: Union[str, None] = None,
    ) -> Tuple[Union[None, StudyResponse], Union[None, str]]:
        study_id = study_id.upper().strip("/") if study_id else ""
        if not user_api_token:
            user_api_token, error = self.get_api_token()
        headers = {}
        if user_api_token:
            headers["User-Token"] = user_api_token
        else:
            return None, error

        sub_path = f"/studies/{study_id}/files/tree"
        parameters = {
            "location": "study",
            "include_sub_dir": False,
            "include_internal_files": False,
        }
        paths = [study_id.strip("/")]
        if subdirectory:
            parameters["directory"] = subdirectory.strip("/")
            paths.append(subdirectory.strip("/"))
        url = f"{self.rest_api_base_url.rstrip('/')}/{sub_path.lstrip('/')}"
        data, error = rest_api_get(
            url, timeout=timeout, parameters=parameters, headers=headers
        )
        if data:
            studies_response = StudyResponse.model_validate(data)
            studies_response.study.sort(key=lambda x: x.file)
            return studies_response, None
        else:
            return None, error

    def rebuild_study_folder_content(
        self, study_id: str, folder_index_file_path: Union[str, None] = None
    ) -> Tuple[Union[None, StudyFolderMetadata], List[GenericMessage]]:
        return self.get_study_folder_content(
            study_id=study_id,
            folder_index_file_path=folder_index_file_path,
            rebuild_folder_index_file=True,
        )

    def get_study_folder_content(
        self,
        study_id: str,
        folder_index_file_path: Union[str, None] = None,
        rebuild_folder_index_file: bool = False,
    ) -> Tuple[Union[None, StudyFolderMetadata], List[GenericMessage]]:

        if not study_id:
            return None, [
                GenericMessage(type=GenericMessageType.ERROR, short="Invalid study_id")
            ]

        study_id = study_id.upper().strip("/")

        collector = FtpFolderMetadataCollector(
            client=self.ftp_client,
            study_id=study_id,
            folder_index_file_path=folder_index_file_path,
            rebuild_folder_index_file=rebuild_folder_index_file,
        )

        metadata, messages = collector.get_folder_metadata(study_path=None)

        return metadata, messages



import re
from pathlib import Path

home = str(Path.home())
default_ftp_server_url = "ftp.ebi.ac.uk"
default_remote_repository_root_directory = "/pub/databases/metabolights/studies/public"
default_local_repority_root_path = f"{home}/metabolights_data/studies/data"
default_local_submission_root_path = f"{home}/metabolights_data/submission/data"
default_local_submission_cache_path = f"{home}/metabolights_data/submission/cache"
default_local_submission_credentials_file_path = (
    f"{home}/metabolights_data/submission/credentials/.login"
)
default_local_repority_cache_path = f"{home}/metabolights_data/studies/cache"
default_rest_api_url = "https://www.ebi.ac.uk/metabolights/ws"
default_validation_api_url = "https://www.ebi.ac.uk/metabolights/ws3"
default_study_search_rest_api_url = "https://www.ebi.ac.uk/metabolights/ws3"
default_private_ftp_server_url = "ftp-private.ebi.ac.uk"

IGNORED_FILE_PATTERNS = {r"^AUDIT_FILES(/|$)(.*)", r"^INTERNAL_FILES(/|$)(.*)"}

SKIP_FOLDER_CONTENT_PATTERNS = {
    r"^FILES/.+\.d(/|$)(.*)",
    r"^FILES/.+\.raw(/|$)(.*)",
    r"^FILES/.+\.m(/|$)(.*)",
}

ignore_file_patterns = [re.compile(x, re.IGNORECASE) for x in IGNORED_FILE_PATTERNS]
skip_folder_content_patterns = [
    re.compile(x, re.IGNORECASE) for x in SKIP_FOLDER_CONTENT_PATTERNS
]

COMPRESSED_FILE_EXTENSIONS = [
    ".zip",
    ".zipx",
    ".gz",
    ".tar",
    ".7z",
    ".z",
    ".g7z",
    ".arj",
    ".rar",
    ".bz2",
    ".arj",
    ".z",
    ".war",
]


COMPRESSED_FILE_PATTERNS = {r"^[\w].*" + f"({'|'.join(COMPRESSED_FILE_EXTENSIONS)})$"}

TAG_PATTERNS = {
    "type:hidden_file": {r"^\..*$"},
    "type:isatab_file/assay": {r"^a_[\w].*\.txt$"},
    "type:isatab_file/samples": {r"^s_[\w].*\.txt$"},
    "type:isatab_file/investigation": {r"^i_[\w].*\.txt$"},
    "type:isatab_file/assignments": {r"^m_[\w].*\.tsv$"},
    "type:compressed_file": COMPRESSED_FILE_PATTERNS,
    "type:internal_file": {
        r"^validation_files.json.*$",
        r"^validation_report.json.*$",
        r"^metexplore_mapping.json.*$",
        r"^missing_files.txt.*$",
        r"^files-all.json.*",
    },
}



import pathlib
import uuid
from typing import Dict

from metabolights_utils.tsv import model as actions
from metabolights_utils.tsv.actions.base import BaseTsvAction


class UpdateRowsTsvAction(BaseTsvAction):
    def apply_action(
        self,
        source_file_path: pathlib.Path,
        target_file_path: pathlib.Path,
        action: actions.TsvUpdateRowsAction,
        read_encoding: str = "utf-8",
        write_encoding: str = "utf-8",
    ) -> actions.TsvActionResult:
        result: actions.TsvActionResult = actions.TsvActionResult(action=action)
        if action.action_type != actions.TsvActionType.UPDATE_ROW_DATA:
            result.message = "Action name is not valid"
            return result

        action: actions.TsvUpdateRowsAction = action

        row_data: Dict[int, actions.TsvRowData] = action.rows if action.rows else {}
        if not row_data:
            result.message = "There is not row data"
            return result

        row_indices = list(row_data.keys())
        row_indices.sort()

        if not action.id:
            uuid_value = str(uuid.uuid4().hex)
            action.id = uuid_value

        try:
            with open(source_file_path, "r", encoding=read_encoding) as source:
                header_line = source.readline()
                header_names = header_line.strip("\n").split("\t")
                empty_row = [""] * len(header_names)

                with open(target_file_path, "w", encoding=write_encoding) as target:
                    self.write_row(target, header_names)
                    row_index = -1
                    for line in source:
                        row_index += 1
                        if row_index in row_indices:
                            input_row = row_data[row_index]
                            new_row = self.get_updated_row(empty_row, input_row)
                            self.write_row(target, new_row)
                            row_indices.remove(row_index)
                        else:
                            target.write(line)

                    if len(row_indices):
                        result.message = f"Invalid row indices {', '.join([str(x) for x in row_indices])}"
                        return result
            result.success = True
        except Exception as exc:
            result.message = f"{str(exc)}"

        return result
